# Copyright (c) 2017 Veritas Technologies LLC.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Cinder Driver for HyperScale
"""

import os

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
import six

from cinder import exception
from cinder.i18n import _
from cinder.image import image_utils
from cinder import interface
from cinder import utils
from cinder.volume import driver
from cinder.volume.drivers.veritas import hs_constants as constants
from cinder.volume.drivers.veritas import utils as util

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
TYPE_EPISODIC_SNAP = '0'
TYPE_USER_SNAP = '1'
TYPE_WORKFLOW_SNAP = '2'

BLOCK_SIZE = 8
MAX_REPLICAS = 2
DEFAULT_REPLICAS = 1
POOL_NAME = '{30c39970-ad80-4950-5490-8431abfaaaf0}'
HYPERSCALE_VERSION = '1.0.0'
PROVIDER_LOCATION_MNT = "/hyperscale"
PROVIDER_LOCATION = 'hyperscale-sv:' + PROVIDER_LOCATION_MNT


@interface.volumedriver
class HyperScaleDriver(driver.VolumeDriver):

    VERSION = '1.0'
    # ThirdPartySytems wiki page
    CI_WIKI_NAME = "Veritas_HyperScale_CI"

    def __init__(self, *args, **kwargs):
        """Initialization"""

        super(HyperScaleDriver, self).__init__(*args, **kwargs)

        self.compute_map = {}
        self.vsa_map = {}
        self.compute_meta_map = {}
        self.vsa_compute_map = {}
        self.old_total = 0
        self.old_free = 0
        self.my_dnid = None

    @staticmethod
    def _fetch_config_for_controller():
        return HyperScaleDriver._fetch_config_information(
            persona='controller')

    @staticmethod
    def _fetch_config_for_compute():
        return HyperScaleDriver._fetch_config_information(
            persona='compute')

    @staticmethod
    def _fetch_config_for_datanode():
        return HyperScaleDriver._fetch_config_information(
            persona='datanode')

    @staticmethod
    def _fetch_config_information(persona):
        # Get hyperscale config information for persona
        configuration = util.get_configuration(persona)
        return configuration

    @utils.trace_method
    def check_for_setup_error(self):
        # check if HyperScale has been installed correctly
        try:
            version = util.get_hyperscale_version()

            if version != HYPERSCALE_VERSION:
                raise exception.VolumeBackendAPIException(
                    data=(_("Unsupported version: %s") % version))
        except (exception.ErrorInHyperScaleVersion,
                exception.UnableToExecuteHyperScaleCmd):
            err_msg = _('Exception in getting HyperScale version')
            LOG.exception(err_msg)
            raise exception.VolumeBackendAPIException(data=err_msg)

    def _get_replicas(self, volume, metadata):
        """Get the replicas."""
        try:
            ref_targets = self._get_volume_metadata_value(metadata,
                                                          'reflection_targets')
            if ref_targets is not None:
                replicas = MAX_REPLICAS
            else:
                replicas = DEFAULT_REPLICAS

        except Exception:
            LOG.exception("Exception in getting reflection targets")
            replicas = DEFAULT_REPLICAS

        LOG.debug("Number of replicas: %s", replicas)
        return replicas

    @utils.trace_method
    def do_setup(self, context):
        """Any initialization the volume driver does while starting."""
        super(HyperScaleDriver, self).do_setup(context)

        try:
            # Get computes info
            computes = HyperScaleDriver._fetch_config_for_compute()
            if computes is None:
                computes = {}

            for compute in computes.keys():
                if 'disabled' in computes[compute].keys():
                    disabled = computes[compute]['disabled']
                    if disabled == "1":
                        continue
                vsa_ip = computes[compute]['vsa_ip']
                vsa_isolated_ip = computes[compute]['vsa_isolated_ip']
                vsa_section_header = computes[compute]['vsa_section_header']
                compute_name = computes[compute]['compute_name']
                self.compute_map[vsa_ip] = vsa_isolated_ip
                self.vsa_map[vsa_ip] = vsa_section_header
                self.compute_meta_map[compute_name] = vsa_ip
                self.vsa_compute_map[vsa_ip] = compute_name

            # Get controller info
            cntr_info = HyperScaleDriver._fetch_config_for_controller()
            if cntr_info is None:
                cntr_info = {}

            # Get data node info
            self.my_dnid = util.get_datanode_id()
            datanodes = HyperScaleDriver._fetch_config_for_datanode()
            if datanodes is None:
                datanodes = {}

            for key, value in datanodes.items():
                if self.my_dnid == value['hypervisor_id']:
                    self.datanode_hostname = value['datanode_name']
                    self.datanode_ip = value['data_ip']
                    self.dn_routing_key = value['hypervisor_id']

            LOG.debug("In init compute_map %s", self.compute_map)
            LOG.debug("In init vsa_map %s", self.vsa_map)
            LOG.debug("In init compute_meta_map %s", self.compute_meta_map)

        except (exception.UnableToProcessHyperScaleCmdOutput,
                exception.ErrorInFetchingConfiguration):
            err_msg = _("Unable to initialise the Veritas cinder driver")
            LOG.exception(err_msg)
            raise exception.VolumeBackendAPIException(data=err_msg)

        except Exception:
            err_msg = _("Internal error occurred")
            LOG.exception(err_msg)
            raise exception.VolumeBackendAPIException(data=err_msg)

    @utils.trace_method
    def create_cloned_volume(self, volume, src_vref):
        """Creates a clone of the specified volume."""

        LOG.debug("Clone volume")
        model_update = {}
        try:
            LOG.debug("Clone new volume %(t_id)s from source volume %(s_id)s",
                      {"t_id": volume['id'], "s_id": src_vref['id']})
            # 1. Make a call to DN
            # Check if current_dn_owner is set.

            rt_key = None
            # Get metadata for volume
            metadata = self._get_volume_metadata(src_vref)
            rt_key = self._get_volume_metadata_value(metadata,
                                                     'current_dn_owner')
            if rt_key is None:
                rt_key = self.dn_routing_key

            util.message_data_plane(
                rt_key,
                'hyperscale.storage.dm.volume.clone',
                pool_name=POOL_NAME,
                display_name=util.get_guid_with_curly_brackets(
                    volume['id']),
                version_name=util.get_guid_with_curly_brackets(
                    src_vref['id']),
                volume_raw_size=volume['size'],
                volume_qos=1,
                parent_volume_guid=util.get_guid_with_curly_brackets(
                    src_vref['id']),
                user_id=util.get_guid_with_curly_brackets(
                    volume['user_id']),
                project_id=util.get_guid_with_curly_brackets(
                    volume['project_id']),
                volume_guid=util.get_guid_with_curly_brackets(
                    volume['id']))

            LOG.debug("Volume cloned successfully on data node")

            # Get metadata for volume
            volume_metadata = self._get_volume_metadata(volume)
            parent_cur_dn = self._get_volume_metadata_value(metadata,
                                                            'current_dn_ip')

            metadata_update = {}
            metadata_update['Primary_datanode_ip'] = parent_cur_dn
            metadata_update['current_dn_owner'] = rt_key
            metadata_update['current_dn_ip'] = parent_cur_dn
            metadata_update['source_volid'] = src_vref['id']
            metadata_update['size'] = src_vref['size']

            # 2. Choose a potential replica here.
            # The actual decision to have potential replica is made in NOVA.
            rt_key, rt_dn_ip = self._select_rt(volume,
                                               volume_metadata,
                                               only_select=True)

            if rt_key and rt_dn_ip:
                metadata_update['Potential_secondary_key'] = rt_key
                metadata_update['Potential_secondary_ip'] = rt_dn_ip

        except (exception.UnableToExecuteHyperScaleCmd,
                exception.UnableToProcessHyperScaleCmdOutput):
            with excutils.save_and_reraise_exception():
                LOG.exception('Exception in clone volume', exc_info=True)
        except exception.InvalidMetadataType:
            with excutils.save_and_reraise_exception():
                LOG.exception('Exception updating metadata in clone'
                              ' volume', exc_info=True)

        volume_metadata.update(metadata_update)
        volume['provider_location'] = PROVIDER_LOCATION
        model_update = {'provider_location': volume['provider_location'],
                        'metadata': volume_metadata}

        return model_update

    def _get_datanodes_info(self):
        # Get hyperscale datanode config information from controller

        msg_body = {}
        data = None

        try:
            cmd_out, cmd_error = util.message_controller(
                constants.HS_CONTROLLER_EXCH,
                'hyperscale.controller.get.membership',
                **msg_body)
            LOG.debug("Response Message from Controller: %s",
                      cmd_out)
            payload = cmd_out.get('payload')
            data = payload.get('of_membership')

        except (exception.UnableToExecuteHyperScaleCmd,
                exception.UnableToProcessHyperScaleCmdOutput):
            with excutils.save_and_reraise_exception():
                LOG.exception("Failed to get datanode config "
                              "information from controller")

        return data

    def _select_rt(self, volume, metadata, only_select=False):

        # For the boot vdisk(first vdisk) of the instance, choose any
        # reflection target other than this. For the data disks,
        # retain the reflection target.
        # It will be passed by the caller after reading it from instance
        # metadata.

        LOG.debug("_select_rt ")
        rt_key = self._get_volume_metadata_value(metadata,
                                                 'Secondary_datanode_key')
        rt_dn_ip = self._get_volume_metadata_value(metadata,
                                                   'Secondary_datanode_ip')
        current_dn_ip = self._get_volume_metadata_value(metadata,
                                                        'current_dn_ip')

        if current_dn_ip is not None and rt_dn_ip == current_dn_ip:
            return None, None

        if rt_key is not None and rt_dn_ip is not None:
            return rt_key, rt_dn_ip

        rt_key = 'NA'
        rt_dn_ip = 'NA'
        datanodes = self._get_datanodes_info()
        LOG.debug("Data nodes: %s", datanodes)

        for key, value in datanodes.items():
            if value['personality'] == 'datanode':
                if self.my_dnid != value['hypervisor_id']:
                    LOG.debug("reflection target hypervisor_id: %s",
                              value['hypervisor_id'])
                    LOG.debug("my hypervisor_id: %s", self.my_dnid)
                    rt_dn_ip = value['data_ip']
                    rt_key = value['hypervisor_id']

        if only_select:
            return rt_key, rt_dn_ip

        return rt_key, rt_dn_ip

    def _create_replica(self, volume, metadata):
        """Create vdisk on peer data node."""

        try:
            reflection_target_ip = None
            rt_routing_key, reflection_target_ip = (
                self._select_rt(volume, metadata))
            LOG.debug("_create_replica %(rt_key)s %(rt_ip)s",
                      {"rt_key": rt_routing_key,
                       "rt_ip": reflection_target_ip})

            metadata_update = {}
            metadata_update['Secondary_datanode_key'] = rt_routing_key
            metadata_update['Secondary_datanode_ip'] = reflection_target_ip

            if rt_routing_key is None or rt_routing_key == 'NA':
                return False, None, metadata_update

            instance_id = self._get_volume_metadata_value(metadata,
                                                          'InstanceId')

            util.message_data_plane(
                rt_routing_key,
                'hyperscale.storage.dm.volume.create',
                pool_name=POOL_NAME,
                volume_guid=util.get_guid_with_curly_brackets(
                    volume['id']),
                display_name=util.get_guid_with_curly_brackets(
                    volume['id']),
                volume_raw_size=volume['size'],
                vm_id=util.get_guid_with_curly_brackets(
                    six.text_type(instance_id)),
                is_reflection_source=0,
                dn_reflection_factor=1,
                reflection_src_ip=self.datanode_ip,
                user_id=util.get_guid_with_curly_brackets(
                    volume['user_id']),
                project_id=util.get_guid_with_curly_brackets(
                    volume['project_id']),
                volume_qos=1)
            # Failure handling TBD.
            ret = True
            LOG.debug("Create volume sent to reflection target data node")

        except (exception.VolumeNotFound,
                exception.UnableToProcessHyperScaleCmdOutput,
                exception.ErrorInSendingMsg):
            LOG.error("Exception in creating replica", exc_info = True)
            metadata_update['Secondary_datanode_key'] = 'NA'
            metadata_update['Secondary_datanode_ip'] = 'NA'
            metadata_update['DN_Resiliency'] = 'degraded'
            ret = False
        return ret, reflection_target_ip, metadata_update

    def _get_volume_details_for_create_volume(self,
                                              reflection_target_ip,
                                              volume,
                                              metadata):

        instance_id = self._get_volume_metadata_value(metadata,
                                                      'InstanceId')
        volume_details = {}
        volume_details['pool_name'] = POOL_NAME
        volume_details['volume_guid'] = (
            util.get_guid_with_curly_brackets(volume['id']))
        volume_details['display_name'] = (
            util.get_guid_with_curly_brackets(volume['id']))
        volume_details['volume_raw_size'] = volume['size']
        volume_details['vm_id'] = util.get_guid_with_curly_brackets(
            six.text_type(instance_id))
        volume_details['user_id'] = util.get_guid_with_curly_brackets(
            volume['user_id'])
        volume_details['project_id'] = (
            util.get_guid_with_curly_brackets(volume['project_id']))
        volume_details['volume_qos'] = 1
        volume_details['dn_reflection_factor'] = 0

        if reflection_target_ip is not None:
            volume_details['is_reflection_source'] = 1
            volume_details['dn_reflection_factor'] = 1
            volume_details['reflection_target_ip'] = reflection_target_ip

        return volume_details

    def _get_volume_metadata(self, volume):
        volume_metadata = {}
        if 'volume_metadata' in volume:
            for metadata in volume['volume_metadata']:
                volume_metadata[metadata['key']] = metadata['value']
        return volume_metadata

    def _get_volume_metadata_value(self, metadata, metadata_key):
        metadata_value = None
        if metadata:
            metadata_value = metadata.get(metadata_key)

        LOG.debug("Volume metadata key %(m_key)s, value %(m_val)s",
                  {"m_key": metadata_key, "m_val": metadata_value})
        return metadata_value

    @utils.trace_method
    def create_volume(self, volume):
        """Creates a hyperscale volume."""

        model_update = {}
        metadata_update = {}
        reflection_target_ip = None
        LOG.debug("Create volume")
        try:
            volume_metadata = self._get_volume_metadata(volume)

            # 1. Check how many replicas needs to be created.
            replicas = self._get_replicas(volume, volume_metadata)
            if replicas > 1:
                # 2. Create replica on peer datanode.
                LOG.debug("Create volume message sent to peer data node")
                ret, reflection_target_ip, metadata_update = (
                    self._create_replica(volume, volume_metadata))
                if ret is False:
                    metadata_update['DN_Resiliency'] = 'degraded'
                    # Do not fail volume creation, just create one replica.
                    reflection_target_ip = None

            # 3. Get volume details based on reflection factor
            #    for volume
            volume_details = self._get_volume_details_for_create_volume(
                reflection_target_ip, volume, volume_metadata)

            # 4. Send create volume to data node with volume details
            util.message_data_plane(
                self.dn_routing_key,
                'hyperscale.storage.dm.volume.create',
                **volume_details)
            LOG.debug("Create volume message sent to data node")

            volume_metadata['Primary_datanode_ip'] = self.datanode_ip
            volume_metadata['current_dn_owner'] = self.dn_routing_key
            volume_metadata['current_dn_ip'] = self.datanode_ip
            volume_metadata['hs_image_id'] = util.get_hyperscale_image_id()
            volume_metadata.update(metadata_update)

            volume['provider_location'] = PROVIDER_LOCATION
            model_update = {'provider_location': volume['provider_location'],
                            'metadata': volume_metadata}

        except (exception.UnableToProcessHyperScaleCmdOutput,
                exception.ErrorInSendingMsg):
            with excutils.save_and_reraise_exception():
                LOG.exception('Unable to create hyperscale volume')

        return model_update

    @utils.trace_method
    def delete_volume(self, volume):
        """Deletes a volume."""

        LOG.debug("Delete volume with id %s", volume['id'])
        # 1. Check for provider location
        if not volume['provider_location']:
            LOG.warning('Volume %s does not have provider_location specified',
                        volume['name'])
            raise exception.VolumeMetadataNotFound(
                volume_id=volume['id'],
                metadata_key='provider_location')

        # 2. Message data plane for volume deletion
        message_body = {'display_name': volume['name']}

        # if Secondary_datanode_key is present,
        # delete the replica from secondary datanode.
        rt_key = None

        # Get metadata for volume
        metadata = self._get_volume_metadata(volume)

        rt_key = self._get_volume_metadata_value(metadata,
                                                 'Secondary_datanode_key')
        rt_dn_ip = self._get_volume_metadata_value(metadata,
                                                   'Secondary_datanode_ip')
        current_dn_ip = self._get_volume_metadata_value(metadata,
                                                        'current_dn_ip')
        if current_dn_ip is not None and rt_dn_ip == current_dn_ip:
            rt_key = None

        # Send Delete Volume to Data Node
        try:
            if rt_key is not None:
                util.message_data_plane(
                    rt_key,
                    'hyperscale.storage.dm.volume.delete',
                    **message_body)

            util.message_data_plane(
                self.dn_routing_key,
                'hyperscale.storage.dm.volume.delete',
                **message_body)

        except (exception.UnableToProcessHyperScaleCmdOutput,
                exception.ErrorInSendingMsg):
            LOG.error('Exception while deleting volume', exc_info=True)
            raise exception.VolumeIsBusy(volume_name=volume['name'])

    @utils.trace_method
    def create_snapshot(self, snapshot):
        """Create a snapshot."""

        LOG.debug("Create Snapshot %s", snapshot['volume_id'])
        workflow_id = None
        last_in_eds_seq = None
        model_update = {}
        rt_key = None

        # Get metadata for volume
        snapshot_volume = snapshot.get('volume')
        metadata = snapshot_volume['metadata']
        rt_key = self._get_volume_metadata_value(metadata,
                                                 'current_dn_owner')
        if rt_key is None:
            rt_key = self.dn_routing_key

        # Check for episodic based on metadata key
        workflow_snap = 0

        meta = snapshot.get('metadata')
        LOG.debug('Snapshot metatadata %s', meta)
        if 'SNAPSHOT-COOKIE' in meta.keys():
            snapsize = meta['SIZE']

            # Call DataNode for episodic snapshots
            LOG.debug('Calling Data Node for episodic snapshots')
            message_body = {}
            message_body['snapshot_id'] = (
                util.get_guid_with_curly_brackets(snapshot['id']))
            message_body['volume_guid'] = (
                util.get_guid_with_curly_brackets(
                    snapshot['volume_id']))
            message_body['snapshot_cookie'] = meta['SNAPSHOT-COOKIE']

            try:
                # send message to data node
                util.message_data_plane(
                    rt_key,
                    'hyperscale.storage.dm.volume.snapshot.update',
                    **message_body)

                # Update size via cinder api
                if snapsize is not None:
                    model_update['volume_size'] = snapsize.value

                # Set the episodic type metatdata for filtering purpose
                meta['TYPE'] = TYPE_EPISODIC_SNAP
                meta['status'] = 'available'
                meta['datanode_ip'] = self.datanode_ip

            except (exception.VolumeNotFound,
                    exception.UnableToExecuteHyperScaleCmd,
                    exception.UnableToProcessHyperScaleCmdOutput):
                with excutils.save_and_reraise_exception():
                    LOG.exception('Exception in create snapshot')

            model_update['metadata'] = meta
            return model_update

        else:
            out_meta = util.episodic_snap(meta)
            if out_meta.get('update'):
                meta['TYPE'] = out_meta.get('TYPE')
                meta['status'] = out_meta.get('status')
                meta['datanode_ip'] = self.datanode_ip
                model_update['metadata'] = meta
                return model_update

        if 'workflow_id' in meta.keys():
            workflow_snap = 1
            workflow_id = meta['workflow_id']

        if 'monitor_snap' in meta.keys():
            if int(meta['monitor_snap']) == constants.SNAP_RESTORE_RF:
                last_in_eds_seq = 0
            else:
                last_in_eds_seq = 1

        # If code falls through here then it mean its user initiated snapshots
        try:
            # Get metadata for volume
            vsa_routing_key = None
            snapshot_volume = snapshot.get('volume')
            metadata = snapshot_volume['metadata']
            LOG.debug('Calling Compute Node for user initiated snapshots')
            vsa_ip = self._get_volume_metadata_value(metadata,
                                                     'acting_vdisk_owner')
            if vsa_ip is None:
                vsa_ip = self._get_volume_metadata_value(metadata, 'vsa_ip')

            LOG.debug("Create snap on compute vsa %s", vsa_ip)
            if vsa_ip:
                vsa_routing_key = vsa_ip.replace('.', '')

            message_body = {}
            # Set the parent volume id
            message_body['vdisk_id_str'] = (
                util.get_guid_with_curly_brackets(
                    snapshot['volume_id']))
            # Set the snapshot details
            message_body['snapshot_id_str'] = (
                util.get_guid_with_curly_brackets(snapshot['id']))
            message_body['snapshot_name'] = snapshot['name']

            if workflow_snap == 1:
                message_body['workflow_snapshot'] = 1
            else:
                message_body['user_initiated'] = 1

            if last_in_eds_seq is not None:
                message_body['last_in_eds_seq'] = last_in_eds_seq

            # send message to compute node
            util.message_compute_plane(
                vsa_routing_key,
                'hyperscale.storage.nfs.volume.snapshot.create',
                **message_body)

            # Set the snapshot type to either workflow or user initiated
            # snapshot in metatdata for filtering purpose
            if workflow_snap:
                LOG.debug('__help request for WORKFLOW snapshot')
                meta['TYPE'] = TYPE_WORKFLOW_SNAP
                meta['status'] = 'creating'
                meta['datanode_ip'] = self.datanode_ip
            else:
                LOG.debug('__help request for MANUAL snapshot')
                meta['TYPE'] = TYPE_USER_SNAP
                meta['status'] = 'creating'
                meta['datanode_ip'] = self.datanode_ip

            if workflow_id is not None:
                message_body = {}
                message_body['workflow_id'] = workflow_id
                message_body['skip_upto_sentinel'] = (
                    'hyperscale.vdisk.failback.snapmark_sentinel')

                # send message to controller node
                util.message_controller(
                    constants.HS_CONTROLLER_EXCH,
                    'hyperscale.controller.execute.workflow',
                    **message_body)

        except (exception.VolumeNotFound,
                exception.UnableToExecuteHyperScaleCmd,
                exception.UnableToProcessHyperScaleCmdOutput):
            with excutils.save_and_reraise_exception():
                LOG.exception('Exception in create snapshot')

        model_update['metadata'] = meta
        return model_update

    @utils.trace_method
    def delete_snapshot(self, snapshot):
        """Deletes a snapshot."""

        meta = snapshot.get('metadata')
        if 'force' in meta.keys():
            LOG.debug("Found force flag for snapshot metadata."
                      " Not sending call to datanode ")
            LOG.debug('snapshot metadata %s', meta)
            return

        if 'is_busy' in meta.keys():
            LOG.warning("Snapshot %s is being used, skipping delete",
                        snapshot['id'])
            raise exception.SnapshotIsBusy(snapshot_name=snapshot['id'])
        else:
            LOG.warning("Snapshot %s is being deleted,"
                        " is_busy key not present", snapshot['id'])

        message_body = {}
        message_body['volume_guid'] = (
            util.get_guid_with_curly_brackets(snapshot['volume_id']))
        message_body['snapshot_id'] = (
            util.get_guid_with_curly_brackets(snapshot['id']))

        # HyperScale snapshots whether Episodic or User initiated, all resides
        # in the data plane.
        # Hence delete snapshot operation will go to datanode
        rt_key = None

        # Get metadata for volume
        snapshot_volume = snapshot.get('volume')
        metadata = snapshot_volume['metadata']
        rt_key = self._get_volume_metadata_value(metadata,
                                                 'current_dn_owner')
        if rt_key is None:
            rt_key = self.dn_routing_key

        try:
            # send message to data node
            util.message_data_plane(
                rt_key,
                'hyperscale.storage.dm.version.delete',
                **message_body)

        except (exception.UnableToExecuteHyperScaleCmd,
                exception.UnableToProcessHyperScaleCmdOutput):
            with excutils.save_and_reraise_exception():
                LOG.exception('Exception in delete snapshot')

    @utils.trace_method
    def create_volume_from_snapshot(self, volume, snapshot):
        """Create volume from snapshot."""

        LOG.debug("Create volume from snapshot")
        model_update = {}
        try:
            LOG.debug("Clone new volume %(t_id)s from snapshot with id"
                      " %(s_id)s", {"t_id": volume['id'],
                                    "s_id": volume['snapshot_id']})
            # 1. Make a call to DN
            # Check if current_dn_owner is set.
            # Route the snapshot creation request to current_dn_owner

            rt_key = None

            # Get metadata for volume
            snap_vol = snapshot['volume']
            metadata = snap_vol['metadata']
            rt_key = self._get_volume_metadata_value(metadata,
                                                     'current_dn_owner')
            if rt_key is None:
                rt_key = self.dn_routing_key

            util.message_data_plane(
                rt_key,
                'hyperscale.storage.dm.volume.clone.create',
                pool_name=POOL_NAME,
                display_name=util.get_guid_with_curly_brackets(
                    volume['id']),
                version_name=util.get_guid_with_curly_brackets(
                    volume['snapshot_id']),
                volume_raw_size=volume['size'],
                volume_qos=1,
                parent_volume_guid=util.get_guid_with_curly_brackets(
                    snapshot['volume_id']),
                user_id=util.get_guid_with_curly_brackets(
                    volume['user_id']),
                project_id=util.get_guid_with_curly_brackets(
                    volume['project_id']),
                volume_guid=util.get_guid_with_curly_brackets(
                    volume['id']))

            LOG.debug("Volume created successfully on data node")

            # Get metadata for volume
            volume_metadata = self._get_volume_metadata(volume)
            parent_cur_dn = self._get_volume_metadata_value(metadata,
                                                            'current_dn_ip')

            metadata_update = {}
            metadata_update['snapshot_id'] = snapshot['id']
            metadata_update['parent_volume_guid'] = (
                util.get_guid_with_curly_brackets(
                    snapshot['volume_id']))
            metadata_update['Primary_datanode_ip'] = parent_cur_dn
            metadata_update['current_dn_owner'] = rt_key
            metadata_update['current_dn_ip'] = parent_cur_dn

            # 2. Choose a potential replica here.
            # The actual decision to have potential replica is made in NOVA.
            rt_key, rt_dn_ip = self._select_rt(volume,
                                               volume_metadata,
                                               only_select=True)

            if rt_key and rt_dn_ip:
                metadata_update['Potential_secondary_key'] = rt_key
                metadata_update['Potential_secondary_ip'] = rt_dn_ip

        except (exception.UnableToExecuteHyperScaleCmd,
                exception.UnableToProcessHyperScaleCmdOutput):
            with excutils.save_and_reraise_exception():
                LOG.exception('Exception in creating volume from snapshot')
        except exception.InvalidMetadataType:
            with excutils.save_and_reraise_exception():
                LOG.exception('Exception updating metadata in create'
                              ' volume from snapshot')

        volume_metadata.update(metadata_update)

        volume['provider_location'] = PROVIDER_LOCATION
        model_update = {'provider_location': volume['provider_location'],
                        'metadata': volume_metadata}

        return model_update

    @utils.trace_method
    def get_volume_stats(self, refresh=False):
        """Get volume status."""

        # If 'refresh' is True, run update the stats first.

        LOG.debug("Get volume status")

        self._stats = self._fetch_volume_status()
        new_total = self._stats['total_capacity_gb']
        new_free = self._stats['free_capacity_gb']

        if self.old_total != new_total or self.old_free != new_free:
            self.old_total = new_total
            self.old_free = new_free

            message_body = {'hostname': self.datanode_hostname,
                            'is_admin': 1,
                            'total': new_total,
                            'free': new_free}
            try:
                cmd_out, cmd_error = util.message_controller(
                    constants.HS_CONTROLLER_EXCH,
                    'hyperscale.controller.set.datanode.storage.stats',
                    **message_body)
                LOG.debug("Response Message from Controller: %s",
                          cmd_out)

            except (exception.UnableToExecuteHyperScaleCmd,
                    exception.UnableToProcessHyperScaleCmdOutput):
                with excutils.save_and_reraise_exception():
                    LOG.exception('Exception during fetch stats')

        return self._stats

    @utils.trace_method
    def extend_volume(self, volume, size_gb):
        """Extend volume."""

        LOG.debug("Extend volume")
        try:
            message_body = {}
            message_body['volume_guid'] = (
                util.get_guid_with_curly_brackets(volume['id']))
            message_body['new_size'] = size_gb

            # Send Extend Volume message to Data Node
            util.message_data_plane(
                self.dn_routing_key,
                'hyperscale.storage.dm.volume.extend',
                **message_body)

        except (exception.UnableToProcessHyperScaleCmdOutput,
                exception.ErrorInSendingMsg):
            msg = _('Exception in extend volume %s') % volume['name']
            LOG.exception(msg)
            raise exception.VolumeDriverException(message=msg)

    def _fetch_volume_status(self):
        """Retrieve Volume Stats from Datanode."""

        LOG.debug("Request Volume Stats from Datanode")

        data = {}
        data["volume_backend_name"] = 'Veritas_HyperScale'
        data["vendor_name"] = 'Veritas Technologies LLC'
        data["driver_version"] = self.VERSION
        data["storage_protocol"] = 'nfs'
        data['total_capacity_gb'] = 0.0
        data['free_capacity_gb'] = 0.0
        data['reserved_percentage'] = self.configuration.reserved_percentage
        data['QoS_support'] = False

        try:
            message_body = {}
            # send message to data node
            cmd_out, cmd_error = util.message_data_plane(
                self.dn_routing_key,
                'hyperscale.storage.dm.discover.stats',
                **message_body)

            LOG.debug("Response Message from Datanode: %s", cmd_out)
            payload = cmd_out.get('payload')
            if 'stats' in payload.keys():
                if 'total_capacity' in payload.get(
                        'stats')[0].keys():
                    total_capacity = payload.get(
                        'stats')[0]['total_capacity']

                if 'free_capacity' in payload.get(
                        'stats')[0].keys():
                    free_capacity = payload.get(
                        'stats')[0]['free_capacity']

                if total_capacity is not None:
                    data['total_capacity_gb'] = float(total_capacity)
                    data['free_capacity_gb'] = float(free_capacity)

        except (exception.UnableToExecuteHyperScaleCmd,
                exception.UnableToProcessHyperScaleCmdOutput):
            with excutils.save_and_reraise_exception():
                LOG.exception('Exception during fetch stats')

        return data

    @utils.trace_method
    def initialize_connection(self, volume, connector):
        """Allow connection to connector and return connection info."""
        data = {'export': volume['provider_location'],
                'name': volume['name']}
        return {
            'driver_volume_type': 'veritas_hyperscale',
            'data': data
        }

    def terminate_connection(self, volume, connector, **kwargs):
        """Disallow connection from connector."""
        pass

    def ensure_export(self, ctx, volume):
        """Synchronously recreates an export for a logical volume."""
        pass

    def create_export(self, ctx, volume, connector):

        # Exports the volume. Can optionally return a Dictionary of changes
        # to the volume object to be persisted."""
        pass

    def remove_export(self, ctx, volume):
        """Removes an export for a logical volume."""
        pass

    @utils.trace_method
    def copy_image_to_volume(self, context, volume, image_service, image_id):
        """Fetch the image from image_service and write it to the volume."""

        LOG.debug("copy_image_to_volume volume: %(vol)s "
                  "image service: %(service)s image id: %(id)s.",
                  {'vol': volume,
                   'service': six.text_type(image_service),
                   'id': six.text_type(image_id)})

        path = util.get_image_path(image_id)
        try:
            # Skip image creation if file already exists
            if not os.path.isfile(path):
                image_utils.fetch_to_raw(context,
                                         image_service,
                                         image_id,
                                         path,
                                         BLOCK_SIZE,
                                         size=volume['size'])
            metadata = self._get_volume_metadata(volume)
            hs_img_id = self._get_volume_metadata_value(metadata,
                                                        'hs_image_id')
            util.update_image(path, volume['id'], hs_img_id)

        except (exception.UnableToExecuteHyperScaleCmd,
                exception.UnableToProcessHyperScaleCmdOutput):
            with excutils.save_and_reraise_exception():
                LOG.exception('Failed to copy_image_to_volume')

    @utils.trace_method
    def copy_volume_to_image(self, context, volume, image_service, image_meta):
        """Copy the volume to the specified image."""

        LOG.debug("copy_volume_to_image volume: %(vol)s"
                  " image service:%(service)s image meta: %(meta)s.",
                  {'vol': volume,
                   'service': six.text_type(image_service),
                   'meta': six.text_type(image_meta)})

        try:
            metadata = self._get_volume_metadata(volume)
            hs_img_id = self._get_volume_metadata_value(metadata,
                                                        'hs_image_id')
            path = util.get_image_path(hs_img_id, 'volume')
            image_utils.upload_volume(context,
                                      image_service,
                                      image_meta,
                                      path)

        except (exception.UnableToExecuteHyperScaleCmd,
                exception.UnableToProcessHyperScaleCmdOutput):
            with excutils.save_and_reraise_exception():
                LOG.exception('Failed to copy_volume_to_image')



# Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from oslo_config import cfg
from oslo_utils import timeutils
from sqlalchemy.dialects import mysql
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Index
from sqlalchemy import Integer, MetaData, String, Table, Text, UniqueConstraint

# Get default values via config.  The defaults will either
# come from the default values set in the quota option
# configuration or via cinder.conf if the user has configured
# default values for quotas there.
CONF = cfg.CONF
CONF.import_opt('quota_volumes', 'cinder.quota')
CONF.import_opt('quota_snapshots', 'cinder.quota')
CONF.import_opt('quota_gigabytes', 'cinder.quota')
CONF.import_opt('quota_consistencygroups', 'cinder.quota')

CLASS_NAME = 'default'
CREATED_AT = datetime.datetime.now()  # noqa


def define_tables(meta):
    services = Table(
        'services', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('id', Integer, primary_key=True, nullable=False),
        Column('host', String(255)),
        Column('binary', String(255)),
        Column('topic', String(255)),
        Column('report_count', Integer, nullable=False),
        Column('disabled', Boolean),
        Column('availability_zone', String(255)),
        Column('disabled_reason', String(255)),
        Column('modified_at', DateTime(timezone=False)),
        Column('rpc_current_version', String(36)),
        Column('object_current_version', String(36)),
        Column('replication_status', String(36), default='not-capable'),
        Column('frozen', Boolean, default=False),
        Column('active_backend_id', String(255)),
        Column('cluster_name', String(255), nullable=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    consistencygroups = Table(
        'consistencygroups', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean(create_constraint=True, name=None)),
        Column('id', String(36), primary_key=True, nullable=False),
        Column('user_id', String(255)),
        Column('project_id', String(255)),
        Column('host', String(255)),
        Column('availability_zone', String(255)),
        Column('name', String(255)),
        Column('description', String(255)),
        Column('volume_type_id', String(255)),
        Column('status', String(255)),
        Column('cgsnapshot_id', String(36)),
        Column('source_cgid', String(36)),
        Column('cluster_name', String(255), nullable=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    cgsnapshots = Table(
        'cgsnapshots', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean(create_constraint=True, name=None)),
        Column('id', String(36), primary_key=True, nullable=False),
        Column('consistencygroup_id', String(36),
               ForeignKey('consistencygroups.id'),
               nullable=False),
        Column('user_id', String(255)),
        Column('project_id', String(255)),
        Column('name', String(255)),
        Column('description', String(255)),
        Column('status', String(255)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    groups = Table(
        'groups', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean),
        Column('id', String(36), primary_key=True, nullable=False),
        Column('user_id', String(length=255)),
        Column('project_id', String(length=255)),
        Column('cluster_name', String(255)),
        Column('host', String(length=255)),
        Column('availability_zone', String(length=255)),
        Column('name', String(length=255)),
        Column('description', String(length=255)),
        Column('group_type_id', String(length=36)),
        Column('status', String(length=255)),
        Column('group_snapshot_id', String(36)),
        Column('source_group_id', String(36)),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    group_snapshots = Table(
        'group_snapshots', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean(create_constraint=True, name=None)),
        Column('id', String(36), primary_key=True),
        Column('group_id', String(36),
               ForeignKey('groups.id'),
               nullable=False),
        Column('user_id', String(length=255)),
        Column('project_id', String(length=255)),
        Column('name', String(length=255)),
        Column('description', String(length=255)),
        Column('status', String(length=255)),
        Column('group_type_id', String(length=36)),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    volumes = Table(
        'volumes', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('id', String(36), primary_key=True, nullable=False),
        Column('ec2_id', String(255)),
        Column('user_id', String(255)),
        Column('project_id', String(255)),
        Column('host', String(255)),
        Column('size', Integer),
        Column('availability_zone', String(255)),
        Column('status', String(255)),
        Column('attach_status', String(255)),
        Column('scheduled_at', DateTime),
        Column('launched_at', DateTime),
        Column('terminated_at', DateTime),
        Column('display_name', String(255)),
        Column('display_description', String(255)),
        Column('provider_location', String(256)),
        Column('provider_auth', String(256)),
        Column('snapshot_id', String(36)),
        Column('volume_type_id', String(36)),
        Column('source_volid', String(36)),
        Column('bootable', Boolean),
        Column('provider_geometry', String(255)),
        Column('_name_id', String(36)),
        Column('encryption_key_id', String(36)),
        Column('migration_status', String(255)),
        Column('replication_status', String(255)),
        Column('replication_extended_status', String(255)),
        Column('replication_driver_data', String(255)),
        Column('consistencygroup_id', String(36),
               ForeignKey('consistencygroups.id')),
        Column('provider_id', String(255)),
        Column('multiattach', Boolean),
        Column('previous_status', String(255)),
        Column('cluster_name', String(255), nullable=True),
        Column('group_id', String(36), ForeignKey('groups.id')),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    volume_attachment = Table(
        'volume_attachment', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('id', String(36), primary_key=True, nullable=False),
        Column('volume_id', String(36), ForeignKey('volumes.id'),
               nullable=False),
        Column('attached_host', String(255)),
        Column('instance_uuid', String(36)),
        Column('mountpoint', String(255)),
        Column('attach_time', DateTime),
        Column('detach_time', DateTime),
        Column('attach_mode', String(36)),
        Column('attach_status', String(255)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    snapshots = Table(
        'snapshots', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('id', String(36), primary_key=True, nullable=False),
        Column('volume_id', String(36),
               ForeignKey('volumes.id', name='snapshots_volume_id_fkey'),
               nullable=False),
        Column('user_id', String(255)),
        Column('project_id', String(255)),
        Column('status', String(255)),
        Column('progress', String(255)),
        Column('volume_size', Integer),
        Column('scheduled_at', DateTime),
        Column('display_name', String(255)),
        Column('display_description', String(255)),
        Column('provider_location', String(255)),
        Column('encryption_key_id', String(36)),
        Column('volume_type_id', String(36)),
        Column('cgsnapshot_id', String(36),
               ForeignKey('cgsnapshots.id')),
        Column('provider_id', String(255)),
        Column('provider_auth', String(255)),
        Column('group_snapshot_id', String(36),
               ForeignKey('group_snapshots.id')),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    snapshot_metadata = Table(
        'snapshot_metadata', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('id', Integer, primary_key=True, nullable=False),
        Column('snapshot_id', String(36), ForeignKey('snapshots.id'),
               nullable=False),
        Column('key', String(255)),
        Column('value', String(255)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    quality_of_service_specs = Table(
        'quality_of_service_specs', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean(create_constraint=True, name=None)),
        Column('id', String(36), primary_key=True, nullable=False),
        Column('specs_id', String(36),
               ForeignKey('quality_of_service_specs.id')),
        Column('key', String(255)),
        Column('value', String(255)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    volume_types = Table(
        'volume_types', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('id', String(36), primary_key=True, nullable=False),
        Column('name', String(255)),
        Column('qos_specs_id', String(36),
               ForeignKey('quality_of_service_specs.id')),
        Column('is_public', Boolean),
        Column('description', String(255)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    volume_type_projects = Table(
        'volume_type_projects', meta,
        Column('id', Integer, primary_key=True, nullable=False),
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('volume_type_id', String(36),
               ForeignKey('volume_types.id')),
        Column('project_id', String(255)),
        Column('deleted', Integer),
        UniqueConstraint('volume_type_id', 'project_id', 'deleted'),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    volume_metadata = Table(
        'volume_metadata', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('id', Integer, primary_key=True, nullable=False),
        Column('volume_id', String(36), ForeignKey('volumes.id'),
               nullable=False),
        Column('key', String(255)),
        Column('value', String(255)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    volume_type_extra_specs = Table(
        'volume_type_extra_specs', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('id', Integer, primary_key=True, nullable=False),
        Column('volume_type_id', String(36),
               ForeignKey('volume_types.id',
                          name='volume_type_extra_specs_ibfk_1'),
               nullable=False),
        Column('key', String(255)),
        Column('value', String(255)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    quotas = Table(
        'quotas', meta,
        Column('id', Integer, primary_key=True, nullable=False),
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('project_id', String(255)),
        Column('resource', String(255), nullable=False),
        Column('hard_limit', Integer),
        Column('allocated', Integer, default=0),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    quota_classes = Table(
        'quota_classes', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean(create_constraint=True,
                                  name=None)),
        Column('id', Integer(), primary_key=True),
        Column('class_name', String(255), index=True),
        Column('resource', String(255)),
        Column('hard_limit', Integer(), nullable=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    quota_usages = Table(
        'quota_usages', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean(create_constraint=True,
                                  name=None)),
        Column('id', Integer(), primary_key=True),
        Column('project_id', String(255), index=True),
        Column('resource', String(255)),
        Column('in_use', Integer(), nullable=False),
        Column('reserved', Integer(), nullable=False),
        Column('until_refresh', Integer(), nullable=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    reservations = Table(
        'reservations', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean(create_constraint=True,
                                  name=None)),
        Column('id', Integer(), primary_key=True),
        Column('uuid', String(36), nullable=False),
        Column('usage_id',
               Integer(),
               ForeignKey('quota_usages.id'),
               nullable=True),
        Column('project_id', String(255), index=True),
        Column('resource', String(255)),
        Column('delta', Integer(), nullable=False),
        Column('expire', DateTime(timezone=False)),
        Column('allocated_id', Integer, ForeignKey('quotas.id'),
               nullable=True),
        Index('reservations_deleted_expire_idx',
              'deleted', 'expire'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    volume_glance_metadata = Table(
        'volume_glance_metadata',
        meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean(create_constraint=True, name=None)),
        Column('id', Integer(), primary_key=True, nullable=False),
        Column('volume_id', String(36), ForeignKey('volumes.id')),
        Column('snapshot_id', String(36),
               ForeignKey('snapshots.id')),
        Column('key', String(255)),
        Column('value', Text),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    backups = Table(
        'backups', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean(create_constraint=True, name=None)),
        Column('id', String(36), primary_key=True, nullable=False),
        Column('volume_id', String(36), nullable=False),
        Column('user_id', String(255)),
        Column('project_id', String(255)),
        Column('host', String(255)),
        Column('availability_zone', String(255)),
        Column('display_name', String(255)),
        Column('display_description', String(255)),
        Column('container', String(255)),
        Column('status', String(255)),
        Column('fail_reason', String(255)),
        Column('service_metadata', String(255)),
        Column('service', String(255)),
        Column('size', Integer()),
        Column('object_count', Integer()),
        Column('parent_id', String(36)),
        Column('temp_volume_id', String(36)),
        Column('temp_snapshot_id', String(36)),
        Column('num_dependent_backups', Integer, default=0),
        Column('snapshot_id', String(36)),
        Column('data_timestamp', DateTime),
        Column('restore_volume_id', String(36)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    transfers = Table(
        'transfers', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean),
        Column('id', String(36), primary_key=True, nullable=False),
        Column('volume_id', String(36), ForeignKey('volumes.id'),
               nullable=False),
        Column('display_name', String(255)),
        Column('salt', String(255)),
        Column('crypt_hash', String(255)),
        Column('expires_at', DateTime(timezone=False)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    # Sqlite needs to handle nullable differently
    is_nullable = (meta.bind.name == 'sqlite')

    encryption = Table(
        'encryption', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean(create_constraint=True, name=None)),
        Column('cipher', String(255)),
        Column('control_location', String(255), nullable=is_nullable),
        Column('key_size', Integer),
        Column('provider', String(255), nullable=is_nullable),
        # NOTE(joel-coffman): The volume_type_id must be unique or else the
        # referenced volume type becomes ambiguous. That is, specifying the
        # volume type is not sufficient to identify a particular encryption
        # scheme unless each volume type is associated with at most one
        # encryption scheme.
        Column('volume_type_id', String(36), nullable=is_nullable),
        # NOTE (smcginnis): nullable=True triggers this to not set a default
        # value, but since it's a primary key the resulting schema will end up
        # still being NOT NULL. This is avoiding a case in MySQL where it will
        # otherwise set this to NOT NULL DEFAULT ''. May be harmless, but
        # inconsistent with previous schema.
        Column('encryption_id', String(36), primary_key=True, nullable=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    volume_admin_metadata = Table(
        'volume_admin_metadata', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('id', Integer, primary_key=True, nullable=False),
        Column('volume_id', String(36), ForeignKey('volumes.id'),
               nullable=False),
        Column('key', String(255)),
        Column('value', String(255)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    initiator_data = Table(
        'driver_initiator_data', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('id', Integer, primary_key=True, nullable=False),
        Column('initiator', String(255), index=True, nullable=False),
        Column('namespace', String(255), nullable=False),
        Column('key', String(255), nullable=False),
        Column('value', String(255)),
        UniqueConstraint('initiator', 'namespace', 'key'),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    image_volume_cache = Table(
        'image_volume_cache_entries', meta,
        Column('image_updated_at', DateTime(timezone=False)),
        Column('id', Integer, primary_key=True, nullable=False),
        Column('host', String(255), index=True, nullable=False),
        Column('image_id', String(36), index=True, nullable=False),
        Column('volume_id', String(36), nullable=False),
        Column('size', Integer, nullable=False),
        Column('last_used', DateTime, nullable=False),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    messages = Table(
        'messages', meta,
        Column('id', String(36), primary_key=True, nullable=False),
        Column('project_id', String(36), nullable=False),
        Column('request_id', String(255), nullable=False),
        Column('resource_type', String(36)),
        Column('resource_uuid', String(255), nullable=True),
        Column('event_id', String(255), nullable=False),
        Column('message_level', String(255), nullable=False),
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean),
        Column('expires_at', DateTime(timezone=False)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    cluster = Table(
        'clusters', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean(), default=False),
        Column('id', Integer, primary_key=True, nullable=False),
        Column('name', String(255), nullable=False),
        Column('binary', String(255), nullable=False),
        Column('disabled', Boolean(), default=False),
        Column('disabled_reason', String(255)),
        Column('race_preventer', Integer, nullable=False, default=0),
        # To remove potential races on creation we have a constraint set on
        # name and race_preventer fields, and we set value on creation to 0, so
        # 2 clusters with the same name will fail this constraint.  On deletion
        # we change this field to the same value as the id which will be unique
        # and will not conflict with the creation of another cluster with the
        # same name.
        UniqueConstraint('name', 'binary', 'race_preventer'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    workers = Table(
        'workers', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean(), default=False),
        Column('id', Integer, primary_key=True),
        Column('resource_type', String(40), nullable=False),
        Column('resource_id', String(36), nullable=False),
        Column('status', String(255), nullable=False),
        Column('service_id', Integer, ForeignKey('services.id'),
               nullable=True),
        UniqueConstraint('resource_type', 'resource_id'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    group_types = Table(
        'group_types', meta,
        Column('id', String(36), primary_key=True, nullable=False),
        Column('name', String(255), nullable=False),
        Column('description', String(255)),
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean),
        Column('is_public', Boolean),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    group_type_specs = Table(
        'group_type_specs', meta,
        Column('id', Integer, primary_key=True, nullable=False),
        Column('key', String(255)),
        Column('value', String(255)),
        Column('group_type_id', String(36),
               ForeignKey('group_types.id'),
               nullable=False),
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    group_type_projects = Table(
        'group_type_projects', meta,
        Column('id', Integer, primary_key=True, nullable=False),
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('group_type_id', String(36),
               ForeignKey('group_types.id')),
        Column('project_id', String(length=255)),
        Column('deleted', Boolean(create_constraint=True, name=None)),
        UniqueConstraint('group_type_id', 'project_id', 'deleted'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    grp_vt_mapping = Table(
        'group_volume_type_mapping', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('id', Integer, primary_key=True, nullable=False),
        Column('volume_type_id', String(36), ForeignKey('volume_types.id'),
               nullable=False),
        Column('group_id', String(36),
               ForeignKey('groups.id'), nullable=False),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    return [consistencygroups,
            cgsnapshots,
            groups,
            group_snapshots,
            volumes,
            volume_attachment,
            snapshots,
            snapshot_metadata,
            quality_of_service_specs,
            volume_types,
            volume_type_projects,
            quotas,
            services,
            volume_metadata,
            volume_type_extra_specs,
            quota_classes,
            quota_usages,
            reservations,
            volume_glance_metadata,
            backups,
            transfers,
            encryption,
            volume_admin_metadata,
            initiator_data,
            image_volume_cache,
            messages,
            cluster,
            workers,
            group_types,
            group_type_specs,
            group_type_projects,
            grp_vt_mapping]


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine

    # create all tables
    # Take care on create order for those with FK dependencies
    tables = define_tables(meta)

    for table in tables:
        table.create()

    if migrate_engine.name == "mysql":
        tables = ["consistencygroups",
                  "cgsnapshots",
                  "snapshots",
                  "snapshot_metadata",
                  "quality_of_service_specs",
                  "volume_types",
                  "volume_type_projects",
                  "volumes",
                  "volume_attachment",
                  "migrate_version",
                  "quotas",
                  "services",
                  "volume_metadata",
                  "volume_type_extra_specs",
                  "quota_classes",
                  "quota_usages",
                  "reservations",
                  "volume_glance_metadata",
                  "backups",
                  "transfers",
                  "encryption",
                  "volume_admin_metadata",
                  "driver_initiator_data",
                  "image_volume_cache_entries"]

        migrate_engine.execute("SET foreign_key_checks = 0")
        for table in tables:
            migrate_engine.execute(
                "ALTER TABLE %s CONVERT TO CHARACTER SET utf8" % table)
        migrate_engine.execute("SET foreign_key_checks = 1")
        migrate_engine.execute(
            "ALTER DATABASE %s DEFAULT CHARACTER SET utf8" %
            migrate_engine.url.database)
        migrate_engine.execute("ALTER TABLE %s Engine=InnoDB" % table)

    # Set default quota class values
    quota_classes = Table('quota_classes', meta, autoload=True)
    qci = quota_classes.insert()
    qci.execute({'created_at': CREATED_AT,
                 'class_name': CLASS_NAME,
                 'resource': 'volumes',
                 'hard_limit': CONF.quota_volumes,
                 'deleted': False, })
    # Set default snapshots
    qci.execute({'created_at': CREATED_AT,
                 'class_name': CLASS_NAME,
                 'resource': 'snapshots',
                 'hard_limit': CONF.quota_snapshots,
                 'deleted': False, })
    # Set default gigabytes
    qci.execute({'created_at': CREATED_AT,
                 'class_name': CLASS_NAME,
                 'resource': 'gigabytes',
                 'hard_limit': CONF.quota_gigabytes,
                 'deleted': False, })
    qci.execute({'created_at': CREATED_AT,
                 'class_name': CLASS_NAME,
                 'resource': 'consistencygroups',
                 'hard_limit': CONF.quota_consistencygroups,
                 'deleted': False, })
    qci.execute({'created_at': CREATED_AT,
                 'class_name': CLASS_NAME,
                 'resource': 'per_volume_gigabytes',
                 'hard_limit': -1,
                 'deleted': False, })
    qci.execute({'created_at': CREATED_AT,
                 'class_name': CLASS_NAME,
                 'resource': 'groups',
                 'hard_limit': CONF.quota_groups,
                 'deleted': False, })

    workers = Table('workers', meta, autoload=True)

    # This is only necessary for mysql, and since the table is not in use this
    # will only be a schema update.
    if migrate_engine.name.startswith('mysql'):
        try:
            workers.c.updated_at.alter(mysql.DATETIME(fsp=6))
        except Exception:
            # MySQL v5.5 or earlier don't support sub-second resolution so we
            # may have cleanup races in Active-Active configurations, that's
            # why upgrading is recommended in that case.
            # Code in Cinder is capable of working with 5.5, so for 5.5 there's
            # no problem
            pass

    # TODO(geguileo): Once we remove support for MySQL 5.5 we have to create
    # an upgrade migration to remove this row.
    # Set workers table sub-second support sentinel
    wi = workers.insert()
    now = timeutils.utcnow().replace(microsecond=123)
    wi.execute({'created_at': now,
                'updated_at': now,
                'deleted': False,
                'resource_type': 'SENTINEL',
                'resource_id': 'SUB-SECOND',
                'status': 'OK'})



# Copyright (c) 2016 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.i18n import _
from cinder.volume.drivers.coprhd.helpers import commoncoprhdapi as common


class Tenant(common.CoprHDResource):

    URI_SERVICES_BASE = ''
    URI_TENANT = URI_SERVICES_BASE + '/tenant'
    URI_TENANTS = URI_SERVICES_BASE + '/tenants/{0}'
    URI_TENANTS_SUBTENANT = URI_TENANTS + '/subtenants'

    def tenant_query(self, label):
        """Returns the UID of the tenant specified by the hierarchical name.

        (ex tenant1/tenant2/tenant3)
        """

        if common.is_uri(label):
            return label

        tenant_id = self.tenant_getid()

        if not label:
            return tenant_id

        subtenants = self.tenant_list(tenant_id)
        subtenants.append(self.tenant_show(None))

        for tenant in subtenants:
            if tenant['name'] == label:
                rslt = self.tenant_show_by_uri(tenant['id'])
                if rslt:
                    return tenant['id']

        raise common.CoprHdError(common.CoprHdError.NOT_FOUND_ERR,
                                 (_("Tenant %s: not found") % label))

    def tenant_show(self, label):
        """Returns the details of the tenant based on its name."""
        if label:
            tenant_id = self.tenant_query(label)
        else:
            tenant_id = self.tenant_getid()

        return self.tenant_show_by_uri(tenant_id)

    def tenant_getid(self):
        (s, h) = common.service_json_request(self.ipaddr, self.port,
                                             "GET", Tenant.URI_TENANT, None)

        o = common.json_decode(s)
        return o['id']

    def tenant_list(self, uri=None):
        """Returns all the tenants under a parent tenant.

        :param uri: The parent tenant name
        :returns: JSON payload of tenant list
        """

        if not uri:
            uri = self.tenant_getid()

        tenantdtls = self.tenant_show_by_uri(uri)

        if(tenantdtls and not ('parent_tenant' in tenantdtls and
                               ("id" in tenantdtls['parent_tenant']))):
            (s, h) = common.service_json_request(
                self.ipaddr, self.port,
                "GET", self.URI_TENANTS_SUBTENANT.format(uri), None)

            o = common.json_decode(s)
            return o['subtenant']

        else:
            return []

    def tenant_show_by_uri(self, uri):
        """Makes REST API call to retrieve tenant details based on UUID."""
        (s, h) = common.service_json_request(self.ipaddr, self.port, "GET",
                                             Tenant.URI_TENANTS.format(uri),
                                             None)

        o = common.json_decode(s)
        if 'inactive' in o and o['inactive']:
            return None

        return o

    def get_tenant_by_name(self, tenant):
        uri = None
        if not tenant:
            uri = self.tenant_getid()
        else:
            if not common.is_uri(tenant):
                uri = self.tenant_query(tenant)
            else:
                uri = tenant
            if not uri:
                raise common.CoprHdError(common.CoprHdError.NOT_FOUND_ERR,
                                         (_("Tenant %s: not found") % tenant))
        return uri



# Copyright (c) 2016 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.i18n import _
from cinder.volume.drivers.coprhd.helpers import commoncoprhdapi as common
from cinder.volume.drivers.coprhd.helpers import tenant


class Host(common.CoprHDResource):

    # All URIs for the Host operations
    URI_HOST_DETAILS = "/compute/hosts/{0}"
    URI_HOST_LIST_INITIATORS = "/compute/hosts/{0}/initiators"
    URI_COMPUTE_HOST = "/compute/hosts"

    def query_by_name(self, host_name, tenant_name=None):
        """Search host matching host_name and tenant if tenant_name provided.

        tenant_name is optional
        """
        hostList = self.list_all(tenant_name)
        for host in hostList:
            hostUri = host['id']
            hostDetails = self.show_by_uri(hostUri)
            if hostDetails:
                if hostDetails['name'] == host_name:
                    return hostUri

        raise common.CoprHdError(common.CoprHdError.NOT_FOUND_ERR, (_(
                                 "Host with name: %s not found") % host_name))

    def list_initiators(self, host_name):
        """Lists all initiators for the given host.

        :param host_name: The name of the host
        """
        if not common.is_uri(host_name):
            hostUri = self.query_by_name(host_name, None)
        else:
            hostUri = host_name

        (s, h) = common.service_json_request(
            self.ipaddr, self.port, "GET",
            Host.URI_HOST_LIST_INITIATORS.format(hostUri),
            None)
        o = common.json_decode(s)

        if not o or "initiator" not in o:
            return []

        return common.get_node_value(o, 'initiator')

    def list_all(self, tenant_name):
        """Gets the ids and self links for all compute elements."""
        restapi = self.URI_COMPUTE_HOST
        tenant_obj = tenant.Tenant(self.ipaddr, self.port)
        if tenant_name is None:
            tenant_uri = tenant_obj.tenant_getid()
        else:
            tenant_uri = tenant_obj.tenant_query(tenant_name)
        restapi = restapi + "?tenant=" + tenant_uri

        (s, h) = common.service_json_request(
            self.ipaddr, self.port,
            "GET",
            restapi,
            None)
        o = common.json_decode(s)
        return o['host']

    def show_by_uri(self, uri):
        """Makes REST API call to retrieve Host details based on its UUID."""
        (s, h) = common.service_json_request(self.ipaddr, self.port, "GET",
                                             Host.URI_HOST_DETAILS.format(uri),
                                             None)
        o = common.json_decode(s)
        inactive = common.get_node_value(o, 'inactive')

        if inactive:
            return None
        return o



# Copyright (c) 2014 LINBIT HA Solutions GmbH
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""

This driver connects Cinder to an installed DRBDmanage instance, see
http://drbd.linbit.com/users-guide-9.0/ch-openstack.html
for more details.

"""


import eventlet
import json
import six
import socket
import time
import uuid

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import units


from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder.volume import configuration
from cinder.volume import driver

try:
    import dbus
    import drbdmanage.consts as dm_const
    import drbdmanage.exceptions as dm_exc
    import drbdmanage.utils as dm_utils
except ImportError:
    # Used for the tests, when no DRBDmanage is installed
    dbus = None
    dm_const = None
    dm_exc = None
    dm_utils = None


LOG = logging.getLogger(__name__)

drbd_opts = [
    cfg.IntOpt('drbdmanage_redundancy',
               default=1,
               help='Number of nodes that should replicate the data.'),
    cfg.StrOpt('drbdmanage_resource_policy',
               default='{"ratio": "0.51", "timeout": "60"}',
               help='Resource deployment completion wait policy.'),
    cfg.StrOpt('drbdmanage_disk_options',
               default='{"c-min-rate": "4M"}',
               help='Disk options to set on new resources. '
               'See http://www.drbd.org/en/doc/users-guide-90/re-drbdconf'
               ' for all the details.'),
    cfg.StrOpt('drbdmanage_net_options',
               default='{"connect-int": "4", "allow-two-primaries": "yes", '
               '"ko-count": "30", "max-buffers": "20000", '
               '"ping-timeout": "100"}',
               help='Net options to set on new resources. '
               'See http://www.drbd.org/en/doc/users-guide-90/re-drbdconf'
               ' for all the details.'),
    cfg.StrOpt('drbdmanage_resource_options',
               default='{"auto-promote-timeout": "300"}',
               help='Resource options to set on new resources. '
               'See http://www.drbd.org/en/doc/users-guide-90/re-drbdconf'
               ' for all the details.'),
    cfg.StrOpt('drbdmanage_snapshot_policy',
               default='{"count": "1", "timeout": "60"}',
               help='Snapshot completion wait policy.'),
    cfg.StrOpt('drbdmanage_resize_policy',
               default='{"timeout": "60"}',
               help='Volume resize completion wait policy.'),
    cfg.StrOpt('drbdmanage_resource_plugin',
               default="drbdmanage.plugins.plugins.wait_for.WaitForResource",
               help='Resource deployment completion wait plugin.'),
    cfg.StrOpt('drbdmanage_snapshot_plugin',
               default="drbdmanage.plugins.plugins.wait_for.WaitForSnapshot",
               help='Snapshot completion wait plugin.'),
    cfg.StrOpt('drbdmanage_resize_plugin',
               default="drbdmanage.plugins.plugins.wait_for.WaitForVolumeSize",
               help='Volume resize completion wait plugin.'),
    cfg.BoolOpt('drbdmanage_devs_on_controller',
                default=True,
                help='''If set, the c-vol node will receive a useable
                /dev/drbdX device, even if the actual data is stored on
                other nodes only.
                This is useful for debugging, maintenance, and to be
                able to do the iSCSI export from the c-vol node.''')
    # TODO(PM): offsite_redundancy?
    # TODO(PM): choose DRBDmanage storage pool?
]


CONF = cfg.CONF
CONF.register_opts(drbd_opts, group=configuration.SHARED_CONF_GROUP)


AUX_PROP_CINDER_VOL_ID = "cinder-id"
AUX_PROP_TEMP_CLIENT = "cinder-is-temp-client"
DM_VN_PREFIX = 'CV_'  # sadly 2CV isn't allowed by DRBDmanage
DM_SN_PREFIX = 'SN_'


# Need to be set later, so that the tests can fake
CS_DEPLOYED = None
CS_DISKLESS = None
CS_UPD_CON = None


class DrbdManageBaseDriver(driver.VolumeDriver):
    """Cinder driver that uses DRBDmanage for storage."""

    VERSION = '1.1.0'
    drbdmanage_dbus_name = 'org.drbd.drbdmanaged'
    drbdmanage_dbus_interface = '/interface'

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "Cinder_Jenkins"

    def __init__(self, *args, **kwargs):
        self.empty_list = dbus.Array([], signature="a(s)")
        self.empty_dict = dbus.Array([], signature="a(ss)")

        super(DrbdManageBaseDriver, self).__init__(*args, **kwargs)

        self.configuration.append_config_values(drbd_opts)
        if not self.drbdmanage_dbus_name:
            self.drbdmanage_dbus_name = 'org.drbd.drbdmanaged'
        if not self.drbdmanage_dbus_interface:
            self.drbdmanage_dbus_interface = '/interface'
        self.drbdmanage_redundancy = int(getattr(self.configuration,
                                                 'drbdmanage_redundancy', 1))
        self.drbdmanage_devs_on_controller = bool(
            getattr(self.configuration,
                    'drbdmanage_devs_on_controller',
                    True))
        self.dm_control_vol = ".drbdctrl"

        self.backend_name = self.configuration.safe_get(
            'volume_backend_name') or 'drbdmanage'

        js_decoder = json.JSONDecoder()
        self.policy_resource = js_decoder.decode(
            self.configuration.safe_get('drbdmanage_resource_policy'))
        self.policy_snapshot = js_decoder.decode(
            self.configuration.safe_get('drbdmanage_snapshot_policy'))
        self.policy_resize = js_decoder.decode(
            self.configuration.safe_get('drbdmanage_resize_policy'))

        self.resource_options = js_decoder.decode(
            self.configuration.safe_get('drbdmanage_resource_options'))
        self.net_options = js_decoder.decode(
            self.configuration.safe_get('drbdmanage_net_options'))
        self.disk_options = js_decoder.decode(
            self.configuration.safe_get('drbdmanage_disk_options'))

        self.plugin_resource = self.configuration.safe_get(
            'drbdmanage_resource_plugin')
        self.plugin_snapshot = self.configuration.safe_get(
            'drbdmanage_snapshot_plugin')
        self.plugin_resize = self.configuration.safe_get(
            'drbdmanage_resize_plugin')

        # needed as per pep8:
        #   F841 local variable 'CS_DEPLOYED' is assigned to but never used
        global CS_DEPLOYED, CS_DISKLESS, CS_UPD_CON
        CS_DEPLOYED = dm_const.CSTATE_PREFIX + dm_const.FLAG_DEPLOY
        CS_DISKLESS = dm_const.CSTATE_PREFIX + dm_const.FLAG_DISKLESS
        CS_UPD_CON = dm_const.CSTATE_PREFIX + dm_const.FLAG_UPD_CON

    def dbus_connect(self):
        self.odm = dbus.SystemBus().get_object(self.drbdmanage_dbus_name,
                                               self.drbdmanage_dbus_interface)
        self.odm.ping()

    def call_or_reconnect(self, fn, *args):
        """Call DBUS function; on a disconnect try once to reconnect."""
        try:
            return fn(*args)
        except dbus.DBusException as e:
            LOG.warning("Got disconnected; trying to reconnect. (%s)", e)
            self.dbus_connect()
            # Old function object is invalid, get new one.
            return getattr(self.odm, fn._method_name)(*args)

    def _fetch_answer_data(self, res, key, level=None, req=True):
        for code, fmt, data in res:
            if code == dm_exc.DM_INFO:
                if level and level != fmt:
                    continue

                value = [v for k, v in data if k == key]
                if value:
                    if len(value) == 1:
                        return value[0]
                    else:
                        return value

        if req:
            if level:
                l = level + ":" + key
            else:
                l = key

            msg = _('DRBDmanage driver error: expected key "%s" '
                    'not in answer, wrong DRBDmanage version?') % l
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        return None

    def do_setup(self, context):
        """Any initialization the volume driver does while starting."""
        super(DrbdManageBaseDriver, self).do_setup(context)
        self.dbus_connect()

    def check_for_setup_error(self):
        """Verify that requirements are in place to use DRBDmanage driver."""
        if not all((dbus, dm_exc, dm_const, dm_utils)):
            msg = _('DRBDmanage driver setup error: some required '
                    'libraries (dbus, drbdmanage.*) not found.')
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)
        if self.odm.ping() != 0:
            message = _('Cannot ping DRBDmanage backend')
            raise exception.VolumeBackendAPIException(data=message)

    def _clean_uuid(self):
        """Returns a UUID string, WITHOUT braces."""
        # Some uuid library versions put braces around the result!?
        # We don't want them, just a plain [0-9a-f-]+ string.
        id = str(uuid.uuid4())
        id = id.replace("{", "")
        id = id.replace("}", "")
        return id

    def _check_result(self, res, ignore=None, ret=0):
        seen_success = False
        seen_error = False
        result = ret
        for (code, fmt, arg_l) in res:
            # convert from DBUS to Python
            arg = dict(arg_l)
            if ignore and code in ignore:
                if not result:
                    result = code
                continue
            if code == dm_exc.DM_SUCCESS:
                seen_success = True
                continue
            if code == dm_exc.DM_INFO:
                continue
            seen_error = _("Received error string: %s") % (fmt % arg)

        if seen_error:
            raise exception.VolumeBackendAPIException(data=seen_error)
        if seen_success:
            return ret
        # by default okay - or the ignored error code.
        return ret

    # DRBDmanage works in kiB units; Cinder uses GiB.
    def _vol_size_to_dm(self, size):
        return int(size * units.Gi / units.Ki)

    def _vol_size_to_cinder(self, size):
        return int(size * units.Ki / units.Gi)

    def is_clean_volume_name(self, name, prefix):
        try:
            if (name.startswith(CONF.volume_name_template % "") and
                    uuid.UUID(name[7:]) is not None):
                return prefix + name[7:]
        except ValueError:
            return None

        try:
            if uuid.UUID(name) is not None:
                return prefix + name
        except ValueError:
            return None

    def _call_policy_plugin(self, plugin, pol_base, pol_this):
        """Returns True for done, False for timeout."""

        pol_inp_data = dict(pol_base)
        pol_inp_data.update(pol_this,
                            starttime=str(time.time()))

        retry = 0
        while True:
            res, pol_result = self.call_or_reconnect(
                self.odm.run_external_plugin,
                plugin,
                pol_inp_data)
            self._check_result(res)

            if pol_result['result'] == dm_const.BOOL_TRUE:
                return True

            if pol_result['timeout'] == dm_const.BOOL_TRUE:
                return False

            eventlet.sleep(min(0.5 + retry / 5, 2))
            retry += 1

    def _wait_for_node_assignment(self, res_name, vol_nr, nodenames,
                                  filter_props=None, timeout=90,
                                  check_vol_deployed=True):
        """Return True as soon as one assignment matches the filter."""

        # TODO(LINBIT): unify with policy plugins

        if not filter_props:
            filter_props = self.empty_dict

        end_time = time.time() + timeout

        retry = 0
        while time.time() < end_time:
            res, assgs = self.call_or_reconnect(self.odm.list_assignments,
                                                nodenames, [res_name], 0,
                                                filter_props, self.empty_list)
            self._check_result(res)

            if len(assgs) > 0:
                for assg in assgs:
                    vols = assg[3]

                    for v_nr, v_prop in vols:
                        if (v_nr == vol_nr):
                            if not check_vol_deployed:
                                # no need to check
                                return True

                            if v_prop[CS_DEPLOYED] == dm_const.BOOL_TRUE:
                                return True

            retry += 1
            # Not yet
            LOG.warning('Try #%(try)d: Volume "%(res)s"/%(vol)d '
                        'not yet deployed on "%(host)s", waiting.',
                        {'try': retry, 'host': nodenames,
                         'res': res_name, 'vol': vol_nr})

            eventlet.sleep(min(0.5 + retry / 5, 2))

        # Timeout
        return False

    def _priv_hash_from_volume(self, volume):
        return dm_utils.dict_to_aux_props({
            AUX_PROP_CINDER_VOL_ID: volume['id'],
        })

    def snapshot_name_from_cinder_snapshot(self, snapshot):
        sn_name = self.is_clean_volume_name(snapshot['id'], DM_SN_PREFIX)
        return sn_name

    def _res_and_vl_data_for_volume(self, volume, empty_ok=False):
        """Find DRBD resource and volume ID.

        A DRBD resource might consist of several "volumes"
        (think consistency groups).
        So we have to find the number of the volume within one resource.
        Returns resource name, volume number, and resource
        and volume properties.
        """

        # If we get a string, use it as-is.
        # Else it's a dictionary; then get the ID.
        if isinstance(volume, six.string_types):
            v_uuid = volume
        else:
            v_uuid = volume['id']

        res, rl = self.call_or_reconnect(self.odm.list_volumes,
                                         self.empty_dict,
                                         0,
                                         dm_utils.dict_to_aux_props(
                                             {AUX_PROP_CINDER_VOL_ID: v_uuid}),
                                         self.empty_dict)
        self._check_result(res)

        if not rl:
            if empty_ok:
                LOG.debug("No volume %s found.", v_uuid)
                return None, None, None, None
            raise exception.VolumeBackendAPIException(
                data=_("volume %s not found in drbdmanage") % v_uuid)
        if len(rl) > 1:
            raise exception.VolumeBackendAPIException(
                data=_("multiple resources with name %s found by drbdmanage") %
                v_uuid)

        (r_name, r_props, vols) = rl[0]
        if len(vols) != 1:
            raise exception.VolumeBackendAPIException(
                data=_("not exactly one volume with id %s") %
                v_uuid)

        (v_nr, v_props) = vols[0]

        LOG.debug("volume %(uuid)s is %(res)s/%(nr)d; %(rprop)s, %(vprop)s",
                  {'uuid': v_uuid, 'res': r_name, 'nr': v_nr,
                   'rprop': dict(r_props), 'vprop': dict(v_props)})

        return r_name, v_nr, r_props, v_props

    def _resource_and_snap_data_from_snapshot(self, snapshot, empty_ok=False):
        """Find DRBD resource and snapshot name from the snapshot ID."""
        s_uuid = snapshot['id']
        res, rs = self.call_or_reconnect(self.odm.list_snapshots,
                                         self.empty_dict,
                                         self.empty_dict,
                                         0,
                                         dm_utils.dict_to_aux_props(
                                             {AUX_PROP_CINDER_VOL_ID: s_uuid}),
                                         self.empty_dict)
        self._check_result(res)

        if not rs:
            if empty_ok:
                return None
            else:
                raise exception.VolumeBackendAPIException(
                    data=_("no snapshot with id %s found in drbdmanage") %
                    s_uuid)
        if len(rs) > 1:
            raise exception.VolumeBackendAPIException(
                data=_("multiple resources with snapshot ID %s found") %
                s_uuid)

        (r_name, snaps) = rs[0]
        if len(snaps) != 1:
            raise exception.VolumeBackendAPIException(
                data=_("not exactly one snapshot with id %s") % s_uuid)

        (s_name, s_props) = snaps[0]

        LOG.debug("snapshot %(uuid)s is %(res)s/%(snap)s",
                  {'uuid': s_uuid, 'res': r_name, 'snap': s_name})

        return r_name, s_name, s_props

    def _resource_name_volnr_for_volume(self, volume, empty_ok=False):
        res, vol, __, __ = self._res_and_vl_data_for_volume(volume, empty_ok)
        return res, vol

    def local_path(self, volume):
        d_res_name, d_vol_nr = self._resource_name_volnr_for_volume(volume)

        res, data = self.call_or_reconnect(self.odm.text_query,
                                           [dm_const.TQ_GET_PATH,
                                            d_res_name,
                                            str(d_vol_nr)])
        self._check_result(res)

        if len(data) == 1:
            return data[0]

        message = _('Got bad path information from DRBDmanage! (%s)') % data
        raise exception.VolumeBackendAPIException(data=message)

    def _push_drbd_options(self, d_res_name):
        res_opt = {'resource': d_res_name,
                   'target': 'resource',
                   'type': 'reso'}
        res_opt.update(self.resource_options)
        res = self.call_or_reconnect(self.odm.set_drbdsetup_props, res_opt)
        self._check_result(res)

        res_opt = {'resource': d_res_name,
                   'target': 'resource',
                   'type': 'neto'}
        res_opt.update(self.net_options)
        res = self.call_or_reconnect(self.odm.set_drbdsetup_props, res_opt)
        self._check_result(res)

        res_opt = {'resource': d_res_name,
                   'target': 'resource',
                   'type': 'disko'}
        res_opt.update(self.disk_options)
        res = self.call_or_reconnect(self.odm.set_drbdsetup_props, res_opt)
        self._check_result(res)

    def create_volume(self, volume):
        """Creates a DRBD resource.

        We address it later on via the ID that gets stored
        as a private property.
        """

        # TODO(PM): consistency groups
        d_res_name = self.is_clean_volume_name(volume['id'], DM_VN_PREFIX)

        res = self.call_or_reconnect(self.odm.create_resource,
                                     d_res_name,
                                     self.empty_dict)
        self._check_result(res, ignore=[dm_exc.DM_EEXIST], ret=None)

        self._push_drbd_options(d_res_name)

        # If we get DM_EEXIST, then the volume already exists, eg. because
        # deploy gave an error on a previous try (like ENOSPC).
        # Still, there might or might not be the volume in the resource -
        # we have to check that explicitly.
        (__, drbd_vol) = self._resource_name_volnr_for_volume(volume,
                                                              empty_ok=True)
        if not drbd_vol:
            props = self._priv_hash_from_volume(volume)
            # TODO(PM): properties - redundancy, etc
            res = self.call_or_reconnect(self.odm.create_volume,
                                         d_res_name,
                                         self._vol_size_to_dm(volume['size']),
                                         props)
            self._check_result(res)
            drbd_vol = self._fetch_answer_data(res, dm_const.VOL_ID)

        # If we crashed between create_volume and the deploy call,
        # the volume might be defined but not exist on any server. Oh my.
        res = self.call_or_reconnect(self.odm.auto_deploy,
                                     d_res_name, self.drbdmanage_redundancy,
                                     0, False)
        self._check_result(res)

        okay = self._call_policy_plugin(self.plugin_resource,
                                        self.policy_resource,
                                        dict(resource=d_res_name,
                                             volnr=str(drbd_vol)))
        if not okay:
            message = (_('DRBDmanage timeout waiting for volume creation; '
                         'resource "%(res)s", volume "%(vol)s"') %
                       {'res': d_res_name, 'vol': volume['id']})
            raise exception.VolumeBackendAPIException(data=message)

        if self.drbdmanage_devs_on_controller:
            # TODO(pm): CG
            res = self.call_or_reconnect(self.odm.assign,
                                         socket.gethostname(),
                                         d_res_name,
                                         [(dm_const.FLAG_DISKLESS,
                                           dm_const.BOOL_TRUE)])
            self._check_result(res, ignore=[dm_exc.DM_EEXIST])

        return {}

    def delete_volume(self, volume):
        """Deletes a resource."""
        d_res_name, d_vol_nr = self._resource_name_volnr_for_volume(
            volume,
            empty_ok=True)

        if not d_res_name:
            # OK, already gone.
            return True

        # TODO(PM): check if in use? Ask whether Primary, or just check result?
        res = self.call_or_reconnect(self.odm.remove_volume,
                                     d_res_name, d_vol_nr, False)
        self._check_result(res, ignore=[dm_exc.DM_ENOENT])

        # Ask for volumes in that resource that are not scheduled for deletion.
        res, rl = self.call_or_reconnect(self.odm.list_volumes,
                                         [d_res_name],
                                         0,
                                         [(dm_const.TSTATE_PREFIX +
                                           dm_const.FLAG_REMOVE,
                                           dm_const.BOOL_FALSE)],
                                         self.empty_list)
        self._check_result(res)

        # We expect the _resource_ to be here still (we just got a volnr from
        # it!), so just query the volumes.
        # If the resource has no volumes anymore, the current DRBDmanage
        # version (errorneously, IMO) returns no *resource*, too.
        if len(rl) > 1:
            message = _('DRBDmanage expected one resource ("%(res)s"), '
                        'got %(n)d') % {'res': d_res_name, 'n': len(rl)}
            raise exception.VolumeBackendAPIException(data=message)

        # Delete resource, if empty
        if (not rl) or (not rl[0]) or not rl[0][2]:
            res = self.call_or_reconnect(self.odm.remove_resource,
                                         d_res_name, False)
            self._check_result(res, ignore=[dm_exc.DM_ENOENT])

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot."""

        LOG.debug("create vol from snap: from %(snap)s make %(vol)s",
                  {'snap': snapshot['id'], 'vol': volume['id']})
        # TODO(PM): Consistency groups.
        d_res_name, sname, sprop = self._resource_and_snap_data_from_snapshot(
            snapshot)

        new_res = self.is_clean_volume_name(volume['id'], DM_VN_PREFIX)

        r_props = self.empty_dict
        # TODO(PM): consistency groups => different volume number possible
        new_vol_nr = 0
        v_props = [(new_vol_nr, self._priv_hash_from_volume(volume))]

        res = self.call_or_reconnect(self.odm.restore_snapshot,
                                     new_res,
                                     d_res_name,
                                     sname,
                                     r_props,
                                     v_props)
        self._check_result(res, ignore=[dm_exc.DM_ENOENT])

        self._push_drbd_options(d_res_name)

        # TODO(PM): CG
        okay = self._call_policy_plugin(self.plugin_resource,
                                        self.policy_resource,
                                        dict(resource=new_res,
                                             volnr=str(new_vol_nr)))
        if not okay:
            message = (_('DRBDmanage timeout waiting for new volume '
                         'after snapshot restore; '
                         'resource "%(res)s", volume "%(vol)s"') %
                       {'res': new_res, 'vol': volume['id']})
            raise exception.VolumeBackendAPIException(data=message)

        if (('size' in volume) and (volume['size'] > snapshot['volume_size'])):
            LOG.debug("resize volume '%(dst_vol)s' from %(src_size)d to "
                      "%(dst_size)d",
                      {'dst_vol': volume['id'],
                       'src_size': snapshot['volume_size'],
                       'dst_size': volume['size']})
            self.extend_volume(volume, volume['size'])

    def create_cloned_volume(self, volume, src_vref):
        temp_id = self._clean_uuid()
        snapshot = {'id': temp_id}

        self.create_snapshot({'id': temp_id,
                              'volume_id': src_vref['id']})

        snapshot['volume_size'] = src_vref['size']
        self.create_volume_from_snapshot(volume, snapshot)

        self.delete_snapshot(snapshot)

    def _update_volume_stats(self):
        data = {}

        data["vendor_name"] = 'Open Source'
        data["driver_version"] = self.VERSION
        # This has to match the name set in the cinder volume driver spec,
        # so keep it lowercase
        data["volume_backend_name"] = self.backend_name
        data["pools"] = []

        res, free, total = self.call_or_reconnect(self.odm.cluster_free_query,
                                                  self.drbdmanage_redundancy)
        self._check_result(res)

        location_info = ('DrbdManageBaseDriver:%(cvol)s:%(dbus)s' %
                         {'cvol': self.dm_control_vol,
                          'dbus': self.drbdmanage_dbus_name})

        # add volumes
        res, rl = self.call_or_reconnect(self.odm.list_volumes,
                                         self.empty_list,
                                         0,
                                         self.empty_dict,
                                         self.empty_list)
        self._check_result(res)
        total_volumes = 0
        for res in rl:
            total_volumes += len(res[2])

        # TODO(PM): multiple DRBDmanage instances and/or multiple pools
        single_pool = {}
        single_pool.update(dict(
            pool_name=data["volume_backend_name"],
            free_capacity_gb=self._vol_size_to_cinder(free),
            total_capacity_gb=self._vol_size_to_cinder(total),
            reserved_percentage=self.configuration.reserved_percentage,
            location_info=location_info,
            total_volumes=total_volumes,
            filter_function=self.get_filter_function(),
            goodness_function=self.get_goodness_function(),
            QoS_support=False))

        data["pools"].append(single_pool)

        self._stats = data
        return self._stats

    def extend_volume(self, volume, new_size):
        d_res_name, d_vol_nr = self._resource_name_volnr_for_volume(volume)

        res = self.call_or_reconnect(self.odm.resize_volume,
                                     d_res_name, d_vol_nr, -1,
                                     self._vol_size_to_dm(new_size),
                                     0)
        self._check_result(res)

        okay = self._call_policy_plugin(self.plugin_resize,
                                        self.policy_resize,
                                        dict(resource=d_res_name,
                                             volnr=str(d_vol_nr),
                                             req_size=str(new_size)))
        if not okay:
            message = (_('DRBDmanage timeout waiting for volume size; '
                         'volume ID "%(id)s" (res "%(res)s", vnr %(vnr)d)') %
                       {'id': volume['id'],
                        'res': d_res_name, 'vnr': d_vol_nr})
            raise exception.VolumeBackendAPIException(data=message)

    def create_snapshot(self, snapshot):
        """Creates a snapshot."""
        sn_name = self.snapshot_name_from_cinder_snapshot(snapshot)

        d_res_name, d_vol_nr = self._resource_name_volnr_for_volume(
            snapshot["volume_id"])

        res, data = self.call_or_reconnect(self.odm.list_assignments,
                                           self.empty_dict,
                                           [d_res_name],
                                           0,
                                           {CS_DISKLESS: dm_const.BOOL_FALSE},
                                           self.empty_list)
        self._check_result(res)

        nodes = [d[0] for d in data]
        if len(nodes) < 1:
            raise exception.VolumeBackendAPIException(
                _('Snapshot res "%s" that is not deployed anywhere?') %
                (d_res_name))

        props = self._priv_hash_from_volume(snapshot)
        res = self.call_or_reconnect(self.odm.create_snapshot,
                                     d_res_name, sn_name, nodes, props)
        self._check_result(res)

        okay = self._call_policy_plugin(self.plugin_snapshot,
                                        self.policy_snapshot,
                                        dict(resource=d_res_name,
                                             snapshot=sn_name))
        if not okay:
            message = (_('DRBDmanage timeout waiting for snapshot creation; '
                         'resource "%(res)s", snapshot "%(sn)s"') %
                       {'res': d_res_name, 'sn': sn_name})
            raise exception.VolumeBackendAPIException(data=message)

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot."""

        d_res_name, sname, _ = self._resource_and_snap_data_from_snapshot(
            snapshot, empty_ok=True)

        if not d_res_name:
            # resource already gone?
            LOG.warning("snapshot: %s not found, "
                        "skipping delete operation", snapshot['id'])
            LOG.info('Successfully deleted snapshot: %s', snapshot['id'])
            return True

        res = self.call_or_reconnect(self.odm.remove_snapshot,
                                     d_res_name, sname, True)
        return self._check_result(res, ignore=[dm_exc.DM_ENOENT])


# Class with iSCSI interface methods
@interface.volumedriver
class DrbdManageIscsiDriver(DrbdManageBaseDriver):
    """Cinder driver that uses the iSCSI protocol. """

    def __init__(self, *args, **kwargs):
        super(DrbdManageIscsiDriver, self).__init__(*args, **kwargs)
        target_driver = self.target_mapping[
            self.configuration.safe_get('target_helper')]

        LOG.debug('Attempting to initialize DRBD driver with the '
                  'following target_driver: %s',
                  target_driver)

        self.target_driver = importutils.import_object(
            target_driver,
            configuration=self.configuration,
            db=self.db,
            executor=self._execute)

    def get_volume_stats(self, refresh=False):
        """Get volume status."""

        self._update_volume_stats()
        self._stats["storage_protocol"] = "iSCSI"
        return self._stats

    def ensure_export(self, context, volume):
        volume_path = self.local_path(volume)
        return self.target_driver.ensure_export(
            context,
            volume,
            volume_path)

    def create_export(self, context, volume, connector):
        volume_path = self.local_path(volume)
        export_info = self.target_driver.create_export(
            context,
            volume,
            volume_path)

        return {'provider_location': export_info['location'],
                'provider_auth': export_info['auth'], }

    def remove_export(self, context, volume):
        return self.target_driver.remove_export(context, volume)

    def initialize_connection(self, volume, connector):
        return self.target_driver.initialize_connection(volume, connector)

    def validate_connector(self, connector):
        return self.target_driver.validate_connector(connector)

    def terminate_connection(self, volume, connector, **kwargs):
        return self.target_driver.terminate_connection(volume,
                                                       connector,
                                                       **kwargs)

# for backwards compatibility keep the old class name, too
DrbdManageDriver = DrbdManageIscsiDriver


# Class with DRBD transport mode
@interface.volumedriver
class DrbdManageDrbdDriver(DrbdManageBaseDriver):
    """Cinder driver that uses the DRBD protocol. """

    def __init__(self, *args, **kwargs):
        super(DrbdManageDrbdDriver, self).__init__(*args, **kwargs)

    def get_volume_stats(self, refresh=False):
        """Get volume status."""

        self._update_volume_stats()
        self._stats["storage_protocol"] = "DRBD"
        return self._stats

    def _return_local_access(self, nodename, volume,
                             d_res_name=None, volume_path=None):

        if not volume_path:
            volume_path = self.local_path(volume)

        return {
            'driver_volume_type': 'local',
            'data': {
                "device_path": volume_path
            }
        }

    def _return_drbdadm_config(self, volume, nodename,
                               d_res_name=None, volume_path=None):

        if not d_res_name:
            d_res_name, d_vol_nr = self._resource_name_volnr_for_volume(volume)

        res, data = self.call_or_reconnect(
            self.odm.text_query,
            ['export_conf_split_up', nodename, d_res_name])
        self._check_result(res)

        config = six.text_type(data.pop(0))
        subst_data = {}
        while len(data):
            k = data.pop(0)
            subst_data[k] = data.pop(0)

        if not volume_path:
            volume_path = self.local_path(volume)

        return {
            'driver_volume_type': 'drbd',
            'data': {
                'provider_location': ' '.join('drbd', nodename),
                'device': volume_path,
                # TODO(pm): consistency groups
                'devices': [volume_path],
                'provider_auth': subst_data['shared-secret'],
                'config': config,
                'name': d_res_name,
            }
        }

    def _is_external_node(self, nodename):
        """Return whether the given node is an "external" node."""

        # If the node accessing the data (the "initiator" in iSCSI speak,
        # "client" or "target" otherwise) is marked as an FLAG_EXTERNAL
        # node, it does not have DRBDmanage active - and that means
        # we have to send the necessary DRBD configuration.
        #
        # If DRBDmanage is running there, just pushing the (client)
        # assignment is enough to make the local path available.

        res, nodes = self.call_or_reconnect(self.odm.list_nodes,
                                            [nodename], 0,
                                            self.empty_dict,
                                            [dm_const.FLAG_EXTERNAL])
        self._check_result(res)

        if len(nodes) != 1:
            msg = _('Expected exactly one node called "%s"') % nodename
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        __, nodeattr = nodes[0]

        return getattr(nodeattr, dm_const.FLAG_EXTERNAL,
                       dm_const.BOOL_FALSE) == dm_const.BOOL_TRUE

    def _return_connection_data(self, nodename, volume, d_res_name=None):
        if nodename and self._is_external_node(nodename):
            return self._return_drbdadm_config(nodename,
                                               volume,
                                               d_res_name=d_res_name)
        else:
            return self._return_local_access(nodename, volume)

    def create_export(self, context, volume, connector):
        d_res_name, d_vol_nr = self._resource_name_volnr_for_volume(volume)

        nodename = connector["host"]

        # Ensure the node is known to DRBDmanage.
        # Note that this does *not* mean that DRBDmanage has to
        # be installed on it!
        # This is just so that DRBD allows the IP to connect.
        node_prop = {
            dm_const.NODE_ADDR: connector["ip"],
            dm_const.FLAG_DRBDCTRL: dm_const.BOOL_FALSE,
            dm_const.FLAG_STORAGE: dm_const.BOOL_FALSE,
            dm_const.FLAG_EXTERNAL: dm_const.BOOL_TRUE,
        }
        res = self.call_or_reconnect(
            self.odm.create_node, nodename, node_prop)
        self._check_result(res, ignore=[dm_exc.DM_EEXIST])

        # Ensure the data is accessible, by creating an assignment.
        assg_prop = {
            dm_const.FLAG_DISKLESS: dm_const.BOOL_TRUE,
        }
        # If we create the assignment here, it's temporary -
        # and has to be removed later on again.
        assg_prop.update(dm_utils.aux_props_to_dict({
            AUX_PROP_TEMP_CLIENT: dm_const.BOOL_TRUE,
        }))

        res = self.call_or_reconnect(
            self.odm.assign, nodename, d_res_name, assg_prop)
        self._check_result(res, ignore=[dm_exc.DM_EEXIST])

        # Wait for DRBDmanage to have completed that action.

        # A DRBDmanage controlled node will set the cstate:deploy flag;
        # an external node will not be available to change it, so we have
        # to wait for the storage nodes to remove the upd_con flag
        # (ie. they're now ready to receive the connection).
        if self._is_external_node(nodename):
            self._wait_for_node_assignment(
                d_res_name, d_vol_nr, [],
                check_vol_deployed=False,
                filter_props={
                    # must be deployed
                    CS_DEPLOYED: dm_const.BOOL_TRUE,
                    # must be a storage node (not diskless),
                    CS_DISKLESS: dm_const.BOOL_FALSE,
                    # connection must be available, no need for updating
                    CS_UPD_CON: dm_const.BOOL_FALSE,
                })
        else:
            self._wait_for_node_assignment(
                d_res_name, d_vol_nr, [nodename],
                check_vol_deployed=True,
                filter_props={
                    CS_DEPLOYED: dm_const.BOOL_TRUE,
                })

        return self._return_connection_data(nodename, volume)

    def ensure_export(self, context, volume):
        p_location = volume['provider_location']
        if p_location:
            fields = p_location.split(" ")
            nodename = fields[1]
        else:
            nodename = None

        return self._return_connection_data(nodename, volume)

    def initialize_connection(self, volume, connector):

        nodename = connector["host"]

        return self._return_connection_data(nodename, volume)

    def terminate_connection(self, volume, connector,
                             force=False, **kwargs):
        d_res_name, d_vol_nr = self._resource_name_volnr_for_volume(
            volume, empty_ok=True)
        if not d_res_name:
            return

        nodename = connector["host"]

        # If the DRBD volume is diskless on that node, we remove it;
        # if it has local storage, we keep it.
        res, data = self.call_or_reconnect(
            self.odm.list_assignments,
            [nodename], [d_res_name], 0,
            self.empty_list, self.empty_list)
        self._check_result(res, ignore=[dm_exc.DM_ENOENT])

        if len(data) < 1:
            # already removed?!
            LOG.info('DRBD connection for %s already removed',
                     volume['id'])
        elif len(data) == 1:
            __, __, props, __ = data[0]
            my_props = dm_utils.dict_to_aux_props(props)
            diskless = getattr(props,
                               dm_const.FLAG_DISKLESS,
                               dm_const.BOOL_FALSE)
            temp_cli = getattr(my_props,
                               AUX_PROP_TEMP_CLIENT,
                               dm_const.BOOL_FALSE)
            # If diskless assigned,
            if ((diskless == dm_const.BOOL_TRUE) and
                    (temp_cli == dm_const.BOOL_TRUE)):
                # remove the assignment

                # TODO(pm): does it make sense to relay "force" here?
                #           What are the semantics?

                # TODO(pm): consistency groups shouldn't really
                #           remove until *all* volumes are detached

                res = self.call_or_reconnect(self.odm.unassign,
                                             nodename, d_res_name, force)
                self._check_result(res, ignore=[dm_exc.DM_ENOENT])
        else:
            # more than one assignment?
            LOG.error("DRBDmanage: too many assignments returned.")
        return

    def remove_export(self, context, volume):
        pass



# Copyright 2016 Nexenta Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Unit tests for OpenStack Cinder volume driver
"""

import mock
from mock import patch

from cinder import context
from cinder import db
from cinder import test
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit.fake_volume import fake_volume_obj
from cinder.volume import configuration as conf
from cinder.volume.drivers.nexenta.ns5 import jsonrpc
from cinder.volume.drivers.nexenta.ns5 import nfs


class TestNexentaNfsDriver(test.TestCase):
    TEST_SHARE = 'host1:/pool/share'
    TEST_SHARE2_OPTIONS = '-o intr'
    TEST_FILE_NAME = 'test.txt'
    TEST_SHARES_CONFIG_FILE = '/etc/cinder/nexenta-shares.conf'
    TEST_SNAPSHOT_NAME = 'snapshot1'
    TEST_VOLUME_NAME = 'volume1'
    TEST_VOLUME_NAME2 = 'volume2'

    TEST_VOLUME = fake_volume_obj(None, **{
        'name': TEST_VOLUME_NAME,
        'id': fake.VOLUME_ID,
        'size': 1,
        'status': 'available',
        'provider_location': TEST_SHARE
    })

    TEST_VOLUME2 = fake_volume_obj(None, **{
        'name': TEST_VOLUME_NAME2,
        'size': 2,
        'id': fake.VOLUME2_ID,
        'status': 'in-use'
    })

    TEST_SNAPSHOT = {
        'name': TEST_SNAPSHOT_NAME,
        'volume_name': TEST_VOLUME_NAME,
        'volume_size': 1,
        'volume_id': fake.VOLUME_ID
    }

    TEST_SHARE_SVC = 'svc:/network/nfs/server:default'

    def setUp(self):
        super(TestNexentaNfsDriver, self).setUp()
        self.ctxt = context.get_admin_context()
        self.cfg = mock.Mock(spec=conf.Configuration)
        self.cfg.nexenta_dataset_description = ''
        self.cfg.nexenta_mount_point_base = '$state_path/mnt'
        self.cfg.nexenta_sparsed_volumes = True
        self.cfg.nexenta_dataset_compression = 'on'
        self.cfg.nexenta_dataset_dedup = 'off'
        self.cfg.nfs_mount_point_base = '/mnt/test'
        self.cfg.nfs_mount_attempts = 3
        self.cfg.nfs_mount_options = None
        self.cfg.nas_mount_options = 'vers=4'
        self.cfg.reserved_percentage = 20
        self.cfg.nexenta_use_https = False
        self.cfg.nexenta_rest_port = 0
        self.cfg.nexenta_user = 'user'
        self.cfg.nexenta_password = 'pass'
        self.cfg.max_over_subscription_ratio = 20.0
        self.cfg.nas_host = '1.1.1.1'
        self.cfg.nas_share_path = 'pool/share'
        self.nef_mock = mock.Mock()
        self.mock_object(jsonrpc, 'NexentaJSONProxy',
                         lambda *_, **__: self.nef_mock)
        self.drv = nfs.NexentaNfsDriver(configuration=self.cfg)
        self.drv.db = db
        self.drv.do_setup(self.ctxt)

    def _create_volume_db_entry(self):
        vol = {
            'id': fake.VOLUME_ID,
            'size': 1,
            'status': 'available',
            'provider_location': self.TEST_SHARE
        }
        return db.volume_create(self.ctxt, vol)['id']

    def test_check_for_setup_error(self):
        self.nef_mock.get.return_value = {'data': []}
        self.assertRaises(
            LookupError, lambda: self.drv.check_for_setup_error())

    def test_initialize_connection(self):
        data = {
            'export': self.TEST_VOLUME['provider_location'], 'name': 'volume'}
        self.assertEqual({
            'driver_volume_type': self.drv.driver_volume_type,
            'data': data
        }, self.drv.initialize_connection(self.TEST_VOLUME, None))

    @patch('cinder.volume.drivers.nexenta.ns5.nfs.'
           'NexentaNfsDriver._create_regular_file')
    @patch('cinder.volume.drivers.nexenta.ns5.nfs.'
           'NexentaNfsDriver._create_sparsed_file')
    @patch('cinder.volume.drivers.nexenta.ns5.nfs.'
           'NexentaNfsDriver._ensure_share_mounted')
    @patch('cinder.volume.drivers.nexenta.ns5.nfs.'
           'NexentaNfsDriver._share_folder')
    def test_do_create_volume(self, share, ensure, sparsed, regular):
        ensure.return_value = True
        share.return_value = True
        self.nef_mock.get.return_value = 'on'
        self.drv._do_create_volume(self.TEST_VOLUME)

        url = 'storage/pools/pool/filesystems'
        data = {
            'name': 'share/volume-' + fake.VOLUME_ID,
            'compressionMode': 'on',
            'dedupMode': 'off',
        }
        self.nef_mock.post.assert_called_with(url, data)

    @patch('cinder.volume.drivers.nexenta.ns5.nfs.'
           'NexentaNfsDriver._ensure_share_mounted')
    def test_delete_volume(self, ensure):
        self._create_volume_db_entry()
        self.nef_mock.get.return_value = {}
        self.drv.delete_volume(self.TEST_VOLUME)
        self.nef_mock.delete.assert_called_with(
            'storage/pools/pool/filesystems/share%2Fvolume-' +
            fake.VOLUME_ID + '?snapshots=true')

    def test_create_snapshot(self):
        self._create_volume_db_entry()
        self.drv.create_snapshot(self.TEST_SNAPSHOT)
        url = ('storage/pools/pool/filesystems/share%2Fvolume-' +
               fake.VOLUME_ID + '/snapshots')
        data = {'name': self.TEST_SNAPSHOT['name']}
        self.nef_mock.post.assert_called_with(url, data)

    def test_delete_snapshot(self):
        self._create_volume_db_entry()
        self.drv.delete_snapshot(self.TEST_SNAPSHOT)
        url = ('storage/pools/pool/filesystems/share%2Fvolume-' +
               fake.VOLUME_ID + '/snapshots/snapshot1')
        self.drv.delete_snapshot(self.TEST_SNAPSHOT)
        self.nef_mock.delete.assert_called_with(url)

    @patch('cinder.volume.drivers.nexenta.ns5.nfs.'
           'NexentaNfsDriver.extend_volume')
    @patch('cinder.volume.drivers.nexenta.ns5.nfs.'
           'NexentaNfsDriver.local_path')
    @patch('cinder.volume.drivers.nexenta.ns5.nfs.'
           'NexentaNfsDriver._share_folder')
    def test_create_volume_from_snapshot(self, share, path, extend):
        self._create_volume_db_entry()
        url = ('storage/pools/%(pool)s/'
               'filesystems/%(fs)s/snapshots/%(snap)s/clone') % {
            'pool': 'pool',
            'fs': '%2F'.join(['share', 'volume-' + fake.VOLUME_ID]),
            'snap': self.TEST_SNAPSHOT['name']
        }
        path = '/'.join(['pool/share', self.TEST_VOLUME2['name']])
        data = {'targetPath': path}
        self.drv.create_volume_from_snapshot(
            self.TEST_VOLUME2, self.TEST_SNAPSHOT)
        self.nef_mock.post.assert_called_with(url, data)

        # make sure the volume get extended!
        extend.assert_called_once_with(self.TEST_VOLUME2, 2)

    @patch('cinder.volume.drivers.nexenta.ns5.nfs.'
           'NexentaNfsDriver.local_path')
    @patch('oslo_concurrency.processutils.execute')
    def test_extend_volume_sparsed(self, _execute, path):
        self._create_volume_db_entry()
        path.return_value = 'path'

        self.drv.extend_volume(self.TEST_VOLUME, 2)

        _execute.assert_called_with(
            'truncate', '-s', '2G',
            'path',
            root_helper='sudo cinder-rootwrap /etc/cinder/rootwrap.conf',
            run_as_root=True)

    @patch('cinder.volume.drivers.nexenta.ns5.nfs.'
           'NexentaNfsDriver.local_path')
    @patch('oslo_concurrency.processutils.execute')
    def test_extend_volume_nonsparsed(self, _execute, path):
        self._create_volume_db_entry()
        path.return_value = 'path'
        with mock.patch.object(self.drv,
                               'sparsed_volumes',
                               False):

            self.drv.extend_volume(self.TEST_VOLUME, 2)

            _execute.assert_called_with(
                'dd', 'if=/dev/zero', 'seek=1073741824',
                'of=path',
                'bs=1M', 'count=1024',
                root_helper='sudo cinder-rootwrap /etc/cinder/rootwrap.conf',
                run_as_root=True)

    def test_get_capacity_info(self):
        self.nef_mock.get.return_value = {
            'bytesAvailable': 1000,
            'bytesUsed': 100}
        self.assertEqual(
            (1000, 900, 100), self.drv._get_capacity_info('pool/share'))



# Copyright (c) 2015 Alex Meade
# Copyright (c) 2015 Yogesh Kshirsagar
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from cinder import test
from cinder.tests.unit.volume.drivers.netapp.eseries import test_driver
import cinder.volume.drivers.netapp.eseries.fc_driver as fc
from cinder.volume.drivers.netapp import utils as na_utils


class NetAppESeriesFibreChannelDriverTestCase(test_driver
                                              .NetAppESeriesDriverTestCase,
                                              test.TestCase):

    PROTOCOL = 'fc'

    @mock.patch.object(na_utils, 'validate_instantiation')
    def test_instantiation(self, mock_validate_instantiation):
        fc.NetAppEseriesFibreChannelDriver(configuration=mock.Mock())

        self.assertTrue(mock_validate_instantiation.called)



# Copyright (c) 2017 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from copy import deepcopy
import datetime
import hashlib
import random
import re
from xml.dom import minidom

from cinder.objects.group import Group
from oslo_log import log as logging
from oslo_utils import strutils
import six

from cinder import exception
from cinder.i18n import _
from cinder.objects import fields
from cinder.volume import volume_types


LOG = logging.getLogger(__name__)
# SHARED CONSTANTS
ISCSI = 'iscsi'
FC = 'fc'
INTERVAL = 'interval'
RETRIES = 'retries'
VOLUME_ELEMENT_NAME_PREFIX = 'OS-'
VMAX_AFA_MODELS = ['VMAX250F', 'VMAX450F', 'VMAX850F', 'VMAX950F']
MAX_SRP_LENGTH = 16
TRUNCATE_5 = 5
TRUNCATE_27 = 27

ARRAY = 'array'
SLO = 'slo'
WORKLOAD = 'workload'
SRP = 'srp'
PORTGROUPNAME = 'storagetype:portgroupname'
DEVICE_ID = 'device_id'
INITIATOR_CHECK = 'initiator_check'
SG_NAME = 'storagegroup_name'
MV_NAME = 'maskingview_name'
IG_NAME = 'init_group_name'
PARENT_SG_NAME = 'parent_sg_name'
CONNECTOR = 'connector'
VOL_NAME = 'volume_name'
EXTRA_SPECS = 'extra_specs'
IS_RE = 'replication_enabled'
DISABLECOMPRESSION = 'storagetype:disablecompression'
REP_SYNC = 'Synchronous'
REP_ASYNC = 'Asynchronous'
REP_METRO = 'Metro'
REP_MODE = 'rep_mode'
RDF_SYNC_STATE = 'synchronized'
RDF_SYNCINPROG_STATE = 'syncinprog'
RDF_CONSISTENT_STATE = 'consistent'
RDF_SUSPENDED_STATE = 'suspended'
RDF_FAILEDOVER_STATE = 'failed over'
RDF_ACTIVE = 'active'
RDF_ACTIVEACTIVE = 'activeactive'
RDF_ACTIVEBIAS = 'activebias'
METROBIAS = 'metro_bias'

# Cinder.conf vmax configuration
VMAX_SERVER_IP = 'san_ip'
VMAX_USER_NAME = 'san_login'
VMAX_PASSWORD = 'san_password'
VMAX_SERVER_PORT = 'san_rest_port'
VMAX_ARRAY = 'vmax_array'
VMAX_WORKLOAD = 'vmax_workload'
VMAX_SRP = 'vmax_srp'
VMAX_SERVICE_LEVEL = 'vmax_service_level'
VMAX_PORT_GROUPS = 'vmax_port_groups'


class VMAXUtils(object):
    """Utility class for Rest based VMAX volume drivers.

    This Utility class is for VMAX volume drivers based on Unisphere Rest API.
    """

    def __init__(self):
        """Utility class for Rest based VMAX volume drivers."""

    def get_host_short_name(self, host_name):
        """Returns the short name for a given qualified host name.

        Checks the host name to see if it is the fully qualified host name
        and returns part before the dot. If there is no dot in the host name
        the full host name is returned.
        :param host_name: the fully qualified host name
        :returns: string -- the short host_name
        """
        host_array = host_name.split('.')
        if len(host_array) > 1:
            short_host_name = host_array[0]
        else:
            short_host_name = host_name

        return self.generate_unique_trunc_host(short_host_name)

    @staticmethod
    def get_volumetype_extra_specs(volume, volume_type_id=None):
        """Gets the extra specs associated with a volume type.

        :param volume: the volume dictionary
        :param volume_type_id: Optional override for volume.volume_type_id
        :returns: dict -- extra_specs - the extra specs
        :raises: VolumeBackendAPIException
        """
        extra_specs = {}

        try:
            if volume_type_id:
                type_id = volume_type_id
            else:
                type_id = volume.volume_type_id
            if type_id is not None:
                extra_specs = volume_types.get_volume_type_extra_specs(type_id)
        except Exception as e:
            LOG.debug('Exception getting volume type extra specs: %(e)s',
                      {'e': six.text_type(e)})
        return extra_specs

    @staticmethod
    def get_short_protocol_type(protocol):
        """Given the protocol type, return I for iscsi and F for fc.

        :param protocol: iscsi or fc
        :returns: string -- 'I' for iscsi or 'F' for fc
        """
        if protocol.lower() == ISCSI.lower():
            return 'I'
        elif protocol.lower() == FC.lower():
            return 'F'
        else:
            return protocol

    @staticmethod
    def truncate_string(str_to_truncate, max_num):
        """Truncate a string by taking first and last characters.

        :param str_to_truncate: the string to be truncated
        :param max_num: the maximum number of characters
        :returns: string -- truncated string or original string
        """
        if len(str_to_truncate) > max_num:
            new_num = len(str_to_truncate) - max_num // 2
            first_chars = str_to_truncate[:max_num // 2]
            last_chars = str_to_truncate[new_num:]
            str_to_truncate = first_chars + last_chars
        return str_to_truncate

    @staticmethod
    def get_time_delta(start_time, end_time):
        """Get the delta between start and end time.

        :param start_time: the start time
        :param end_time: the end time
        :returns: string -- delta in string H:MM:SS
        """
        delta = end_time - start_time
        return six.text_type(datetime.timedelta(seconds=int(delta)))

    def get_default_storage_group_name(
            self, srp_name, slo, workload, is_compression_disabled=False,
            is_re=False, rep_mode=None):
        """Determine default storage group from extra_specs.

        :param srp_name: the name of the srp on the array
        :param slo: the service level string e.g Bronze
        :param workload: the workload string e.g DSS
        :param is_compression_disabled:  flag for disabling compression
        :param is_re: flag for replication
        :param rep_mode: flag to indicate replication mode
        :returns: storage_group_name
        """
        if slo and workload:
            prefix = ("OS-%(srpName)s-%(slo)s-%(workload)s"
                      % {'srpName': srp_name, 'slo': slo,
                         'workload': workload})

            if is_compression_disabled:
                prefix += "-CD"

        else:
            prefix = "OS-no_SLO"
        if is_re:
            prefix += self.get_replication_prefix(rep_mode)

        storage_group_name = ("%(prefix)s-SG" % {'prefix': prefix})
        return storage_group_name

    @staticmethod
    def get_volume_element_name(volume_id):
        """Get volume element name follows naming convention, i.e. 'OS-UUID'.

        :param volume_id: Openstack volume ID containing uuid
        :returns: volume element name in format of OS-UUID
        """
        element_name = volume_id
        uuid_regex = (re.compile(
            '[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}',
            re.I))
        match = uuid_regex.search(volume_id)
        if match:
            volume_uuid = match.group()
            element_name = ("%(prefix)s%(volumeUUID)s"
                            % {'prefix': VOLUME_ELEMENT_NAME_PREFIX,
                               'volumeUUID': volume_uuid})
            LOG.debug(
                "get_volume_element_name elementName:  %(elementName)s.",
                {'elementName': element_name})
        return element_name

    @staticmethod
    def modify_snapshot_prefix(snapshot_name, manage=False, unmanage=False):
        """Modify a Snapshot prefix on VMAX backend.

        Prepare a snapshot name for manage/unmanage snapshot process either
        by adding or removing 'OS-' prefix.

        :param snapshot_name: the old snapshot backend display name
        :param manage: (bool) if the operation is managing a snapshot
        :param unmanage: (bool) if the operation is unmanaging a snapshot
        :return: snapshot name ready for backend VMAX assignment
        """
        new_snap_name = None
        if manage:
            new_snap_name = ("%(prefix)s%(snapshot_name)s"
                             % {'prefix': 'OS-',
                                'snapshot_name': snapshot_name})

        if unmanage:
            snap_split = snapshot_name.split("-", 1)
            if snap_split[0] == 'OS':
                new_snap_name = snap_split[1]

        return new_snap_name

    def generate_unique_trunc_host(self, host_name):
        """Create a unique short host name under 16 characters.

        :param host_name: long host name
        :returns: truncated host name
        """
        if host_name and len(host_name) > 16:
            host_name = host_name.lower()
            m = hashlib.md5()
            m.update(host_name.encode('utf-8'))
            uuid = m.hexdigest()
            new_name = ("%(host)s%(uuid)s"
                        % {'host': host_name[-6:],
                           'uuid': uuid})
            host_name = self.truncate_string(new_name, 16)
        return host_name

    def get_pg_short_name(self, portgroup_name):
        """Create a unique port group name under 12 characters.

        :param portgroup_name: long portgroup_name
        :returns: truncated portgroup_name
        """
        if portgroup_name and len(portgroup_name) > 12:
            portgroup_name = portgroup_name.lower()
            m = hashlib.md5()
            m.update(portgroup_name.encode('utf-8'))
            uuid = m.hexdigest()
            new_name = ("%(pg)s%(uuid)s"
                        % {'pg': portgroup_name[-6:],
                           'uuid': uuid})
            portgroup_name = self.truncate_string(new_name, 12)
        return portgroup_name

    @staticmethod
    def get_default_oversubscription_ratio(max_over_sub_ratio):
        """Override ratio if necessary.

        The over subscription ratio will be overridden if the user supplied
        max oversubscription ratio is less than 1.
        :param max_over_sub_ratio: user supplied over subscription ratio
        :returns: max_over_sub_ratio
        """
        if max_over_sub_ratio < 1.0:
            LOG.info("The user supplied value for max_over_subscription "
                     "ratio is less than 1.0. Using the default value of "
                     "20.0 instead...")
            max_over_sub_ratio = 20.0
        return max_over_sub_ratio

    @staticmethod
    def _process_tag(element, tag_name):
        """Process the tag to get the value.

        :param element: the parent element
        :param tag_name: the tag name
        :returns: nodeValue(can be None)
        """
        node_value = None
        try:
            processed_element = element.getElementsByTagName(tag_name)[0]
            node_value = processed_element.childNodes[0].nodeValue
            if node_value:
                node_value = node_value.strip()
        except IndexError:
            pass
        return node_value

    def _get_connection_info(self, rest_element):
        """Given the filename get the rest server connection details.

        :param rest_element: the rest element
        :returns: dict -- connargs - the connection info dictionary
        :raises: VolumeBackendAPIException
        """
        connargs = {
            'RestServerIp': (
                self._process_tag(rest_element, 'RestServerIp')),
            'RestServerPort': (
                self._process_tag(rest_element, 'RestServerPort')),
            'RestUserName': (
                self._process_tag(rest_element, 'RestUserName')),
            'RestPassword': (
                self._process_tag(rest_element, 'RestPassword'))}

        for k, __ in connargs.items():
            if connargs[k] is None:
                exception_message = (_(
                    "RestServerIp, RestServerPort, RestUserName, "
                    "RestPassword must have valid values."))
                LOG.error(exception_message)
                raise exception.VolumeBackendAPIException(
                    data=exception_message)

        # These can be None
        connargs['SSLCert'] = self._process_tag(rest_element, 'SSLCert')
        connargs['SSLVerify'] = (
            self._process_tag(rest_element, 'SSLVerify'))

        return connargs

    def parse_file_to_get_array_map(self, file_name):
        """Parses a file and gets array map.

        Given a file, parse it to get array and pool(srp).

        .. code:: ini

          <EMC>
          <RestServerIp>10.108.246.202</RestServerIp>
          <RestServerPort>8443</RestServerPort>
          <RestUserName>smc</RestUserName>
          <RestPassword>smc</RestPassword>
          <SSLCert>/path/client.cert</SSLCert>
          <SSLVerify>/path/to/certfile.pem</SSLVerify>
          <PortGroups>
              <PortGroup>OS-PORTGROUP1-PG</PortGroup>
          </PortGroups>
          <Array>000198700439</Array>
          <SRP>SRP_1</SRP>
          </EMC>

        :param file_name: the configuration file
        :returns: list
        """
        LOG.warning("Use of xml file in backend configuration is deprecated "
                    "in Queens and will not be supported in future releases.")
        kwargs = {}
        my_file = open(file_name, 'r')
        data = my_file.read()
        my_file.close()
        dom = minidom.parseString(data)
        try:
            connargs = self._get_connection_info(dom)
            portgroup = self._get_random_portgroup(dom)
            serialnumber = self._process_tag(dom, 'Array')
            if serialnumber is None:
                LOG.error("Array Serial Number must be in the file %(file)s.",
                          {'file': file_name})
            srp_name = self._process_tag(dom, 'SRP')
            if srp_name is None:
                LOG.error("SRP Name must be in the file %(file)s.",
                          {'file': file_name})
            slo = self._process_tag(dom, 'ServiceLevel')
            workload = self._process_tag(dom, 'Workload')
            kwargs = (
                {'RestServerIp': connargs['RestServerIp'],
                 'RestServerPort': connargs['RestServerPort'],
                 'RestUserName': connargs['RestUserName'],
                 'RestPassword': connargs['RestPassword'],
                 'SSLCert': connargs['SSLCert'],
                 'SSLVerify': connargs['SSLVerify'],
                 'SerialNumber': serialnumber,
                 'srpName': srp_name,
                 'PortGroup': portgroup})
            if slo is not None:
                kwargs.update({'ServiceLevel': slo, 'Workload': workload})

        except IndexError:
            pass
        return kwargs

    @staticmethod
    def _get_random_portgroup(element):
        """Randomly choose a portgroup from list of portgroups.

        :param element: the parent element
        :returns: the randomly chosen port group
        """
        portgroupelements = element.getElementsByTagName('PortGroup')
        if portgroupelements and len(portgroupelements) > 0:
            portgroupnames = [portgroupelement.childNodes[0].nodeValue.strip()
                              for portgroupelement in portgroupelements
                              if portgroupelement.childNodes]
            portgroupnames = list(set(filter(None, portgroupnames)))
            pg_len = len(portgroupnames)
            if pg_len > 0:
                return portgroupnames[random.randint(0, pg_len - 1)]
        return None

    def get_temp_snap_name(self, clone_name, source_device_id):
        """Construct a temporary snapshot name for clone operation.

        :param clone_name: the name of the clone
        :param source_device_id: the source device id
        :returns: snap_name
        """
        trunc_clone = self.truncate_string(clone_name, 10)
        snap_name = ("temp-%(device)s-%(clone)s"
                     % {'device': source_device_id, 'clone': trunc_clone})
        return snap_name

    @staticmethod
    def get_array_and_device_id(volume, external_ref):
        """Helper function for manage volume to get array name and device ID.

        :param volume: volume object from API
        :param external_ref: the existing volume object to be manged
        :returns: string value of the array name and device ID
        """
        device_id = external_ref.get(u'source-name', None)
        LOG.debug("External_ref: %(er)s", {'er': external_ref})
        if not device_id:
            device_id = external_ref.get(u'source-id', None)
        host = volume.host
        host_list = host.split('+')
        array = host_list[(len(host_list) - 1)]

        if device_id:
            LOG.debug("Get device ID of existing volume - device ID: "
                      "%(device_id)s, Array: %(array)s.",
                      {'device_id': device_id,
                       'array': array})
        else:
            exception_message = (_("Source volume device ID is required."))
            raise exception.VolumeBackendAPIException(
                data=exception_message)
        return array, device_id

    @staticmethod
    def is_compression_disabled(extra_specs):
        """Check is compression is to be disabled.

        :param extra_specs: extra specifications
        :returns: boolean
        """
        do_disable_compression = False
        if DISABLECOMPRESSION in extra_specs:
            if strutils.bool_from_string(extra_specs[DISABLECOMPRESSION]):
                do_disable_compression = True
        return do_disable_compression

    def change_compression_type(self, is_source_compr_disabled, new_type):
        """Check if volume type have different compression types

        :param is_source_compr_disabled: from source
        :param new_type: from target
        :returns: boolean
        """
        extra_specs = new_type['extra_specs']
        is_target_compr_disabled = self.is_compression_disabled(extra_specs)
        if is_target_compr_disabled == is_source_compr_disabled:
            return False
        else:
            return True

    @staticmethod
    def is_replication_enabled(extra_specs):
        """Check if replication is to be enabled.

        :param extra_specs: extra specifications
        :returns: bool - true if enabled, else false
        """
        replication_enabled = False
        if IS_RE in extra_specs:
            replication_enabled = True
        return replication_enabled

    @staticmethod
    def get_replication_config(rep_device_list):
        """Gather necessary replication configuration info.

        :param rep_device_list: the replication device list from cinder.conf
        :returns: rep_config, replication configuration dict
        """
        rep_config = {}
        if not rep_device_list:
            return None
        else:
            target = rep_device_list[0]
            try:
                rep_config['array'] = target['target_device_id']
                rep_config['srp'] = target['remote_pool']
                rep_config['rdf_group_label'] = target['rdf_group_label']
                rep_config['portgroup'] = target['remote_port_group']

            except KeyError as ke:
                error_message = (_("Failed to retrieve all necessary SRDF "
                                   "information. Error received: %(ke)s.") %
                                 {'ke': six.text_type(ke)})
                LOG.exception(error_message)
                raise exception.VolumeBackendAPIException(data=error_message)

            allow_extend = target.get('allow_extend', 'false')
            if strutils.bool_from_string(allow_extend):
                rep_config['allow_extend'] = True
            else:
                rep_config['allow_extend'] = False

            rep_mode = target.get('mode', '')
            if rep_mode.lower() in ['async', 'asynchronous']:
                rep_config['mode'] = REP_ASYNC
            elif rep_mode.lower() == 'metro':
                rep_config['mode'] = REP_METRO
                metro_bias = target.get('metro_use_bias', 'false')
                if strutils.bool_from_string(metro_bias):
                    rep_config[METROBIAS] = True
                else:
                    rep_config[METROBIAS] = False
                allow_delete_metro = target.get('allow_delete_metro', 'false')
                if strutils.bool_from_string(allow_delete_metro):
                    rep_config['allow_delete_metro'] = True
                else:
                    rep_config['allow_delete_metro'] = False
            else:
                rep_config['mode'] = REP_SYNC

        return rep_config

    @staticmethod
    def is_volume_failed_over(volume):
        """Check if a volume has been failed over.

        :param volume: the volume object
        :returns: bool
        """
        if volume is not None:
            if volume.get('replication_status') and (
                volume.replication_status ==
                    fields.ReplicationStatus.FAILED_OVER):
                    return True
        return False

    @staticmethod
    def update_volume_model_updates(volume_model_updates,
                                    volumes, group_id, status='available'):
        """Update the volume model's status and return it.

        :param volume_model_updates: list of volume model update dicts
        :param volumes: volumes object api
        :param group_id: consistency group id
        :param status: string value reflects the status of the member volume
        :returns: volume_model_updates - updated volumes
        """
        LOG.info("Updating status for group: %(id)s.", {'id': group_id})
        if volumes:
            for volume in volumes:
                volume_model_updates.append({'id': volume.id,
                                             'status': status})
        else:
            LOG.info("No volume found for group: %(cg)s.", {'cg': group_id})
        return volume_model_updates

    @staticmethod
    def get_grp_volume_model_update(volume, volume_dict, group_id):
        """Create and return the volume model update on creation.

        :param volume: volume object
        :param volume_dict: the volume dict
        :param group_id: consistency group id
        :returns: model_update
        """
        LOG.info("Updating status for group: %(id)s.", {'id': group_id})
        model_update = ({'id': volume.id, 'status': 'available',
                         'provider_location': six.text_type(volume_dict)})
        return model_update

    @staticmethod
    def update_extra_specs(extraspecs):
        """Update extra specs.

        :param extraspecs: the additional info
        :returns: extraspecs
        """
        try:
            pool_details = extraspecs['pool_name'].split('+')
            extraspecs[SLO] = pool_details[0]
            if len(pool_details) == 4:
                extraspecs[WORKLOAD] = pool_details[1]
                extraspecs[SRP] = pool_details[2]
                extraspecs[ARRAY] = pool_details[3]
            else:
                # Assume no workload given in pool name
                extraspecs[SRP] = pool_details[1]
                extraspecs[ARRAY] = pool_details[2]
                extraspecs[WORKLOAD] = 'NONE'
        except KeyError:
            LOG.error("Error parsing SLO, workload from"
                      " the provided extra_specs.")
        return extraspecs

    def get_volume_group_utils(self, group, interval, retries):
        """Standard utility for generic volume groups.

        :param group: the generic volume group object to be created
        :param interval: Interval in seconds between retries
        :param retries: Retry count
        :returns: array, intervals_retries_dict
        :raises: VolumeBackendAPIException
        """
        arrays = set()
        # Check if it is a generic volume group instance
        if isinstance(group, Group):
            for volume_type in group.volume_types:
                extra_specs = self.update_extra_specs(volume_type.extra_specs)
                arrays.add(extra_specs[ARRAY])
        else:
            msg = (_("Unable to get volume type ids."))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if len(arrays) != 1:
            if not arrays:
                msg = (_("Failed to get an array associated with "
                         "volume group: %(groupid)s.")
                       % {'groupid': group.id})
            else:
                msg = (_("There are multiple arrays "
                         "associated with volume group: %(groupid)s.")
                       % {'groupid': group.id})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        array = arrays.pop()
        intervals_retries_dict = {INTERVAL: interval, RETRIES: retries}
        return array, intervals_retries_dict

    def update_volume_group_name(self, group):
        """Format id and name consistency group.

        :param group: the generic volume group object
        :returns: group_name -- formatted name + id
        """
        group_name = ""
        if group.name is not None and group.name != group.id:
            group_name = (
                self.truncate_string(
                    group.name, TRUNCATE_27) + "_")

        group_name += group.id
        return group_name

    @staticmethod
    def add_legacy_pools(pools):
        """Add legacy pools to allow extending a volume after upgrade.

        :param pools: the pool list
        :return: pools - the updated pool list
        """
        extra_pools = []
        for pool in pools:
            if 'none' in pool['pool_name'].lower():
                extra_pools.append(pool)
        for pool in extra_pools:
            try:
                slo = pool['pool_name'].split('+')[0]
                srp = pool['pool_name'].split('+')[2]
                array = pool['pool_name'].split('+')[3]
            except IndexError:
                slo = pool['pool_name'].split('+')[0]
                srp = pool['pool_name'].split('+')[1]
                array = pool['pool_name'].split('+')[2]
            new_pool_name = ('%(slo)s+%(srp)s+%(array)s'
                             % {'slo': slo, 'srp': srp, 'array': array})
            new_pool = deepcopy(pool)
            new_pool['pool_name'] = new_pool_name
            pools.append(new_pool)
        return pools

    def check_replication_matched(self, volume, extra_specs):
        """Check volume type and group type.

        This will make sure they do not conflict with each other.

        :param volume: volume to be checked
        :param extra_specs: the extra specifications
        :raises: InvalidInput
        """
        # If volume is not a member of group, skip this check anyway.
        if not volume.group:
            return
        vol_is_re = self.is_replication_enabled(extra_specs)
        group_is_re = volume.group.is_replicated

        if not (vol_is_re == group_is_re):
            msg = _('Replication should be enabled or disabled for both '
                    'volume or group. Volume replication status: '
                    '%(vol_status)s, group replication status: '
                    '%(group_status)s') % {
                        'vol_status': vol_is_re, 'group_status': group_is_re}
            raise exception.InvalidInput(reason=msg)

    @staticmethod
    def check_rep_status_enabled(group):
        """Check replication status for group.

        Group status must be enabled before proceeding with certain
        operations.

        :param group: the group object
        :raises: InvalidInput
        """
        if group.is_replicated:
            if group.replication_status != fields.ReplicationStatus.ENABLED:
                msg = (_('Replication status should be %s for '
                         'replication-enabled group.')
                       % fields.ReplicationStatus.ENABLED)
                LOG.error(msg)
                raise exception.InvalidInput(reason=msg)
        else:
            LOG.debug('Replication is not enabled on group %s, '
                      'skip status check.', group.id)

    @staticmethod
    def get_replication_prefix(rep_mode):
        """Get the replication prefix.

        Replication prefix for storage group naming is based on whether it is
        synchronous, asynchronous, or metro replication mode.

        :param rep_mode: flag to indicate if replication is async
        :return: prefix
        """
        if rep_mode == REP_ASYNC:
            prefix = "-RA"
        elif rep_mode == REP_METRO:
            prefix = "-RM"
        else:
            prefix = "-RE"
        return prefix

    @staticmethod
    def get_async_rdf_managed_grp_name(rep_config):
        """Get the name of the group used for async replication management.

        :param rep_config: the replication configuration
        :return: group name
        """
        async_grp_name = ("OS-%(rdf)s-%(mode)s-rdf-sg"
                          % {'rdf': rep_config['rdf_group_label'],
                             'mode': rep_config['mode']})
        LOG.debug("The async/ metro rdf managed group name is %(name)s",
                  {'name': async_grp_name})
        return async_grp_name

    def is_metro_device(self, rep_config, extra_specs):
        """Determine if a volume is a Metro enabled device.

        :param rep_config: the replication configuration
        :param extra_specs: the extra specifications
        :return: bool
        """
        is_metro = (True if self.is_replication_enabled(extra_specs)
                    and rep_config is not None
                    and rep_config['mode'] == REP_METRO else False)
        return is_metro

    def does_vol_need_rdf_management_group(self, extra_specs):
        """Determine if a volume is a Metro or Async.

        :param extra_specs: the extra specifications
        :return: bool
        """
        if (self.is_replication_enabled(extra_specs) and
                extra_specs.get(REP_MODE, None) in
                [REP_ASYNC, REP_METRO]):
            return True
        return False
